
// Original code:
//   http://www.emit.jp/sha/sha.html

//////////////////////////////////////////////////////////////////////////////
// General Function Library  Copyright (C) 2000 SYN All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////
#include "GflSHA.h"
#include <stdio.h>

static const DWORD c_dwInitH_1_0 = 0x67452301;
static const DWORD c_dwInitH_1_1 = 0xefcdab89;
static const DWORD c_dwInitH_1_2 = 0x98badcfe;
static const DWORD c_dwInitH_1_3 = 0x10325476;
static const DWORD c_dwInitH_1_4 = 0xc3d2e1f0;

static const DWORD c_dwK_1__00_19 = 0x5a827999;
static const DWORD c_dwK_1__20_39 = 0x6ed9eba1;
static const DWORD c_dwK_1__40_59 = 0x8f1bbcdc;
static const DWORD c_dwK_1__60_79 = 0xca62c1d6;


GflSHA1::GflSHA1(void)
{
	Init();
}


GflSHA1::~GflSHA1(void)
{
}


void GflSHA1::Init(void)
{
	m_dwH[0] = c_dwInitH_1_0;
	m_dwH[1] = c_dwInitH_1_1;
	m_dwH[2] = c_dwInitH_1_2;
	m_dwH[3] = c_dwInitH_1_3;
	m_dwH[4] = c_dwInitH_1_4;
	m_dwLNumBits = 0;
	m_dwHNumBits = 0;
	m_nNumChr = 0;
}


void GflSHA1::Load(const BYTE *pBuffer, DWORD dwSize)
{
	if(dwSize == 0) return;

	DWORD dwLNumBits = (m_dwLNumBits + (dwSize << 3));
	if(dwLNumBits < m_dwLNumBits) m_dwHNumBits++;
	m_dwHNumBits += dwSize >> 29;
	m_dwLNumBits = dwLNumBits;

	BYTE *pBlock = (BYTE *)m_aBlock;
	while(dwSize){
		DWORD dwReadSize = (dwSize < SHA_BUFFER_SIZE - m_nNumChr) ? dwSize : (SHA_BUFFER_SIZE - m_nNumChr);
		memcpy(pBlock + m_nNumChr, pBuffer, dwReadSize);
		m_nNumChr += dwReadSize;
		pBuffer += dwReadSize;
		dwSize -= dwReadSize;
		if(m_nNumChr == SHA_BUFFER_SIZE){
			Generate();
			m_nNumChr = 0;
		}
	}
}


void GflSHA1::Final(void)
{
	BYTE cZero = 0x00;
	BYTE cOne  = 0x80;

	DWORD dwHNumBits = ReverseEndian(m_dwHNumBits);
	DWORD dwLNumBits = ReverseEndian(m_dwLNumBits);

	Load(&cOne, 1);
	while(m_nNumChr != SHA_BUFFER_SIZE - 8) Load(&cZero, 1);

	Load((BYTE *)&dwHNumBits, 4);
	Load((BYTE *)&dwLNumBits, 4);
}


string GflSHA1::GetDigest(void)
{
	char s[SHA_HASH * 8 + 1];
	for(int i = 0; i < SHA_HASH; i++) sprintf(s + i * 8, "%08x", m_dwH[i]);
	string strDigest = s;
	return strDigest;
}


void GflSHA1::Generate(void)
{
	int i;
	DWORD Work[SHA_WORK];
	DWORD Hash[SHA_WORK + SHA_HASH];

	for(i = 0; i < SHA_BLOCK; i++) Work[i] = ReverseEndian(m_aBlock[i]);
	for(i = SHA_BLOCK; i < SHA_WORK; i++){
		Work[i] = Rotate(Work[i - 3] ^ Work[i - 8] ^ Work[i - 14] ^ Work[i - 16], 1);
	}

	for(i = 0; i < 5; i++) Hash[SHA_WORK + i] = m_dwH[i];
	DWORD *pHash = &Hash[SHA_WORK];

	for(i = 0; i < 20; i++){
		pHash--;
		pHash[0] = (pHash[2] & (pHash[3] ^ pHash[4])) ^ pHash[4];
		pHash[0] += Rotate(pHash[1], 5) + pHash[5] + Work[i] + c_dwK_1__00_19;
		pHash[2] = Rotate(pHash[2], 30);
	}
	for(i = 20; i < 40; i++){
		pHash--;
		pHash[0] = pHash[2] ^ pHash[3] ^ pHash[4];
		pHash[0] += Rotate(pHash[1], 5) + pHash[5] + Work[i] + c_dwK_1__20_39;
		pHash[2] = Rotate(pHash[2], 30);
	}
	for(i = 40; i < 60; i++){
		pHash--;
		pHash[0] = (pHash[2] & (pHash[3] | pHash[4])) | (pHash[3] & pHash[4]);
		pHash[0] += Rotate(pHash[1], 5) + pHash[5] + Work[i] + c_dwK_1__40_59;
		pHash[2] = Rotate(pHash[2], 30);
	}
	for(i = 60; i < 80; i++){
		pHash--;
		pHash[0] = pHash[2] ^ pHash[3] ^ pHash[4];
		pHash[0] += Rotate(pHash[1], 5) + pHash[5] + Work[i] + c_dwK_1__60_79;
		pHash[2] = Rotate(pHash[2], 30);
	}

	for(i = 0; i < 5; i++) m_dwH[i] += pHash[i];
}

// --------------------
// SHA-256
// --------------------

//////////////////////////////////////////////////////////////////////////////
// General Function Library  Copyright (C) 2000 SYN All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////

static const DWORD c_dwInitH_256[SHA256_HASH] = {
	0x6a09e667, 0xbb67ae85, 0x3c6ef372, 0xa54ff53a, 0x510e527f, 0x9b05688c, 0x1f83d9ab, 0x5be0cd19, 
};

static const DWORD c_dwK_256[SHA256_WORK] = {
	0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5, 
	0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3, 0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174, 
	0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da, 
	0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967, 
	0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13, 0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85, 
	0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070, 
	0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3, 
	0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208, 0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2, 
};


GflSHA256::GflSHA256(void)
{
	Init();
}


GflSHA256::~GflSHA256(void)
{
}


void GflSHA256::Init(void)
{
	for(int i = 0; i < SHA256_HASH; i++) m_dwH[i] = c_dwInitH_256[i];
	m_dwLNumBits = 0;
	m_dwHNumBits = 0;
	m_nNumChr = 0;
}


void GflSHA256::Load(const BYTE *pBuffer, DWORD dwSize)
{
	if(dwSize == 0) return;

	DWORD dwLNumBits = (m_dwLNumBits + (dwSize << 3));
	if(dwLNumBits < m_dwLNumBits) m_dwHNumBits++;
	m_dwHNumBits += dwSize >> 29;
	m_dwLNumBits = dwLNumBits;

	BYTE *pBlock = (BYTE *)m_aBlock;
	while(dwSize){
		DWORD dwReadSize = (dwSize < SHA256_BUFFER_SIZE - m_nNumChr) ? dwSize : (SHA256_BUFFER_SIZE - m_nNumChr);
		memcpy(pBlock + m_nNumChr, pBuffer, dwReadSize);
		m_nNumChr += dwReadSize;
		pBuffer += dwReadSize;
		dwSize -= dwReadSize;
		if(m_nNumChr == SHA256_BUFFER_SIZE){
			Generate();
			m_nNumChr = 0;
		}
	}
}


void GflSHA256::Final(void)
{
	BYTE cZero = 0x00;
	BYTE cOne  = 0x80;

	DWORD dwHNumBits = ReverseEndian(m_dwHNumBits);
	DWORD dwLNumBits = ReverseEndian(m_dwLNumBits);

	Load(&cOne, 1);
	while(m_nNumChr != SHA256_BUFFER_SIZE - 8) Load(&cZero, 1);

	Load((BYTE *)&dwHNumBits, 4);
	Load((BYTE *)&dwLNumBits, 4);
}


string GflSHA256::GetDigest(void)
{
	char s[SHA256_HASH * 8 + 1];
	for(int i = 0; i < SHA256_HASH; i++) sprintf(s + i * 8, "%08x", m_dwH[i]);
	string strDigest = s;
	return strDigest;
}


void GflSHA256::Generate(void)
{
	int i;
	DWORD W[SHA256_WORK];
	DWORD Hash[SHA256_WORK + SHA256_HASH];

	for(i = 0; i < SHA256_BLOCK; i++) W[i] = ReverseEndian(m_aBlock[i]);
	for(i = SHA256_BLOCK; i < SHA256_WORK; i++) W[i] = s1(W[i - 2]) + W[i - 7] + s0(W[i - 15]) + W[i - 16];

	for(i = 0; i < SHA256_HASH; i++) Hash[SHA256_WORK + i] = m_dwH[i];
	DWORD *pHash = &Hash[SHA256_WORK];
	DWORD dwT1, dwT2;
	for(i = 0; i < SHA256_WORK; i++){
		pHash--;
		dwT1 = pHash[8] + S1(pHash[5]) + Ch(pHash[5], pHash[6], pHash[7]) + c_dwK_256[i] + W[i];
		dwT2 = S0(pHash[1]) + Maj(pHash[1], pHash[2], pHash[3]);
		pHash[0] = dwT1 + dwT2;
		pHash[4] += dwT1;
	}
	for(i = 0; i < SHA256_HASH; i++) m_dwH[i] += pHash[i];
}

