//////////////////////////////////////////////////////////////////////////
//
//  hl[Op`
//
//  by pastel
//  http://www.roy.hi-ho.ne.jp/pastel/home/VC/delauney.html
//
//  2001/8/10 first version
//  2005/03/07 TVP Plugin version by mey

#ifndef _TVP_PLUGIN_DELAUNAY_TRIANGULATION_H_
#define _TVP_PLUGIN_DELAUNAY_TRIANGULATION_H_

#include "TVPPlugIn.h"

const int DELAUNAY_INIT_PTSIZE = 5;
const int DELAUNAY_INIT_TRISIZE = 5;
const int DELAUNAY_INIT_STACKSIZE = 5;



class TPTriangle {
public:
	TPTriangle();
	TPTriangle(int ip0, int ip1, int ip2);
	int SetAdjacent(int iEdge, int iAdjacent);
	int GetAdjacent(int iEdge) const;
	void GetPt(int& ip0, int& ip1, int& ip2) const;
	void GetPtPos(int ip, double& px, double& py) const;
	void SetPtPos(int ip, double px, double py);

protected:
	int m_iPt[3];
	int m_iAdjacent[3];
	double m_X[3];
	double m_Y[3];
};


class TPDelaunay {
public:
	TPDelaunay();
	~TPDelaunay();
	void Clear();
	
	void InitPlane(int w, int h);
	void InitQuad(double x0, double y0, double x1, double y1, double x2, double y2, double x3, double y3);

	void SetTrianglePt();

	TPTriangle * GetTriangle(int ip) const;

	// (x, y)𕪊ɉAVȕsD
	// ǉAɎsꍇFALSEԂD
	bool AddPtPos(double x, double y);

	// _̐Ԃ
	int CountPt() const;

	// iPtԖڂ̓_̍W(x, y)ɕԂD
	// iPtɑ݂Ȃ_̔ԍ^ꍇŎsꍇFALSEԂD
	bool GetPtPos(int iPt, double& x, double& y) const;

	// Op`̐Ԃ
	int CountTri() const;

	// iTriԖڂ̎Op`Ȃ3̒_̔ԍԂ
	bool GetPtOfTri(int iTri, int& ip0, int& ip1, int& ip2) const;

	// iTriԖڂ̎Op`Ȃ3̒_̍WԂ
	bool GetPtPosOfTri(int iTri, double& x0, double& y0, double& x1, double& y1, double& x2, double&y2) const;

	// iTri̕iEdgeɑΉӂɗאڂOp`Ԃ
	// Ȃꍇ-1Ԃ
	int GetAdjacent(int iTri, int iEdge) const;

	// iTri0iTri1אڂĂƂD
	// {ɗׂ荇ĂꍇȂD
	// iTri0iTri1ɋʕӂȂꍇ͎sAFALSEԂD
	bool SetAdjacent(int iTri0, int iTri1);

	// (x, y)Ɋ܂ގOp`̔ԍԂ
	// Ȃ-1
	int FindTriContains(double x, double y) const;

	// (x, y)iTriԖڂ̎Op`̒ɂTRUE
	bool IsInside(int iTri, double x, double y) const;

	// iTri0iTri1אڂĂ邩ׂD
	// אڂĂȂTRUE. łȂFALSE.
	//
	// c0 = ʕӂ̔ԍiiTri0猩ꍇj
	// c1 = ʕӂ̔ԍiiTri1猩ꍇj
	// iȂAԂlFALSEꍇc0, c1͈ӖȂj
	bool IsAdjacent(int iTri0, int iTri1, int& c0, int& c1) const;

	// iTri0iTri1̑ΊpւD
	// TRUE, sꍇFALSEԂ
	bool SwapDiagonal(int iTri0, int iTri1);

	// (x0, y0), (x1, y1), (x2, y2)ōOp`̊Oډ~vZA
	// ̒SW(cx, cy)ɁAa̓r2ɕԂD
	int CalcCircumCircle(
		double x0, double y0, double x1, double y1, double x2, double y2,
		double& cx, double& cy, double& r2);

protected:
	class SIntStack {
	public:
		SIntStack();
		~SIntStack();
		int Push(int x);
		bool Pop(int& x);
		int Count() const;
	protected:
		void ClearStack();
		void InitStack();
		bool ExpandStack();
		int *m_pValue;
		int m_Count;
		int m_Capacity;
	};

	// xNg(x0, y0)琳̕ɉ]ʒu(x1, y1)ȂTRUE.
	// 3xNg̊OςgĔ肷D
	bool IsPlusRotation(double x0, double y0, double x1, double y1) const;

	// Ԃl͓o^_̔ԍ
	int AddPtToArray(double x, double y);

	// Ԃl͓o^Op`̔ԍ
	int AddTriToArray(TPTriangle *pTri);

	void ClearPtArray();
	int InitPtArray();
	int ExpandPtArray();

	void ClearTriArray();
	int InitTriArray();
	int ExpandTriArray();

	int m_nPt;
	int m_nPtCapacity;
	double *m_px;
	double *m_py;

	int m_nTri;
	int m_nTriCapacity;
	TPTriangle **m_pTri;

};



#endif _TVP_PLUGIN_DELAUNAY_TRIANGULATION_H_
