
#include "FilterFunc.h"
#include "TPGraphic.h"
#include "TableHolder.h"

// XVtO
extern bool TPFAutoUpdateMode;

// e[uz_[
extern CTableHolder * TableHolder;

typedef TPFUNCCALL tjs_uint32 (*FilterFP)(const tjs_int *, const tjs_int *, const tjs_int *, const tjs_int *, tjs_int);

// G{XtB^p
#define EMBOSS_BASE_COLOR_CODE 128

#define DISTORT_COLOR_CODE_MAX(_cc) if (_cc > 255) _cc = 255;
#define DISTORT_COLOR_CODE_MIN(_cc) if (_cc <   0) _cc = 0;
#define GENERATE_COLOR_CODE_FROM_GS(_cc) \
			_cc *= 0x00010101; \
			_cc |= 0xff000000;

#define FILTER_FUNC_DECL(_fname) \
static TPFUNCCALL tjs_uint32 _fname( \
	const tjs_int * cp1, \
	const tjs_int * cp2, \
	const tjs_int * cp3, \
	const tjs_int * kernel, \
	tjs_int ksum)

#define FILTER_SUM_3X3(_a,_b,_c,_d,_e,_f,_g,_h,_i) \
	((_a) * (cp1[0])) + ((_b) * (cp1[1])) + ((_c) * (cp1[2])) + \
	((_d) * (cp2[0])) + ((_e) * (cp2[1])) + ((_f) * (cp2[2])) + \
	((_g) * (cp3[0])) + ((_h) * (cp3[1])) + ((_i) * (cp3[2]))


#define PRECHECK_DRAW_AREA(_sz) \
	if ((w < _sz) || \
		(h < _sz) || \
		(x + w < draw->left) || \
		(y + h < draw->top) || \
		(x > draw->right) || \
		(y > draw->bottom)) \
		return false; \
	\
	tjs_int left = x; \
	tjs_int top  = y; \
	tjs_int width = w; \
	tjs_int height = h; \
	if (left < draw->left) left = draw->left; \
	if (top  < draw->top)  top  = draw->top; \
	if ((x + w) >= draw->right)  width  = draw->right  - left; \
	if ((y + h) >= draw->bottom) height = draw->bottom - top; \
	if (width < 1 || height < 1) return false;


// ------------------------
// 3x3 p
// J[obt@
// 
#define ALLOC_COLOR_BUFFER_3X3(_b, _c, _len) \
	tjs_int * _b##1 = new tjs_int[_len]; \
	tjs_int * _b##2 = new tjs_int[_len]; \
	tjs_int * _b##3 = new tjs_int[_len]; \
	tjs_int * _c##1 = _b##1; \
	tjs_int * _c##2 = _b##2; \
	tjs_int * _c##3 = _b##3;

// U蕪
#define ASSIGN_COLOR_BUFFER_3X3(_b, _c, _svl) \
		switch (_svl) { \
		case 1: \
			_c##1 = _b##1; \
			_c##2 = _b##2; \
			_c##3 = _b##3; \
			break; \
		case 2: \
			_c##1 = _b##2; \
			_c##2 = _b##3; \
			_c##3 = _b##1; \
			break; \
		case 0: \
		default: \
			_c##1 = _b##3; \
			_c##2 = _b##1; \
			_c##3 = _b##2; \
			break; \
		}

#define COPY_GRAYSCALE_FACTOR_RGB(_lp, _rp, _gp, _bp, _w) \
		CopyGrayScaleFactorOnlyRed_SD(_lp, _rp, _w); \
		CopyGrayScaleFactorOnlyGreen_SD(_lp, _gp, _w); \
		CopyGrayScaleFactorOnlyBlue_SD(_lp, _bp, _w);


// 
#define FREE_COLOR_BUFFER_3X3(_b) \
	delete[] _b##1; \
	delete[] _b##2; \
	delete[] _b##3;




// -------------------------------------------------------------
// ėptB^
// -------------------------------------------------------------
FILTER_FUNC_DECL(genericFilter3x3)
{
	tjs_int c = FILTER_SUM_3X3(
		kernel[0], kernel[1], kernel[2],
		kernel[3], kernel[4], kernel[5],
		kernel[6], kernel[7], kernel[8] );
	DISTORT_COLOR_CODE_MAX(c)
	DISTORT_COLOR_CODE_MIN(c)
	return c;
}

// 
FILTER_FUNC_DECL(genericAvgFilter3x3)
{
	tjs_int c = FILTER_SUM_3X3(
		kernel[0], kernel[1], kernel[2],
		kernel[3], kernel[4], kernel[5],
		kernel[6], kernel[7], kernel[8] );
	c /= ksum;
	DISTORT_COLOR_CODE_MAX(c)
	DISTORT_COLOR_CODE_MIN(c)
	return c;
}



// -------------------------------------------------------------
// fBAtB^
// -------------------------------------------------------------
FILTER_FUNC_DECL(medianFilter3x3)
{
	tjs_int c = FILTER_SUM_3X3(
		 1,  1,  1,
		 1,  1,  1,
		 1,  1,  1 );
	return (c/9);
}



// -------------------------------------------------------------
// SobeltB^iR2L, B2Tj
// -------------------------------------------------------------
FILTER_FUNC_DECL(sobelFilter3x3)
{
	tjs_int h = FILTER_SUM_3X3(
		 -1,  0,  1,
		 -2,  0,  2,
		 -1,  0,  1 );
	tjs_int v = FILTER_SUM_3X3(
		 -1, -2, -1,
		  0,  0,  0,
		  1,  2,  1 );
	tjs_int c = TableHolder->getDistance(h, v);
	DISTORT_COLOR_CODE_MAX(c)
	return c;
}


// -------------------------------------------------------------
// PrewitttB^iR2L, B2Tj
// -------------------------------------------------------------
FILTER_FUNC_DECL(prewittFilter3x3)
{
	tjs_int h = FILTER_SUM_3X3(
		 -1,  0,  1,
		 -1,  0,  1,
		 -1,  0,  1 );
	tjs_int v = FILTER_SUM_3X3(
		 -1, -1, -1,
		  0,  0,  0,
		  1,  1,  1 );
	tjs_int c = TableHolder->getDistance(h, v);
	DISTORT_COLOR_CODE_MAX(c)
	return c;
}


// -------------------------------------------------------------
// RoberttB^ (Robert's Cross operator)
// http://www.incx.nec.co.jp/imap-vision/library/wouter/roberts.html
// -------------------------------------------------------------
FILTER_FUNC_DECL(robertFilter3x3)
{
	tjs_int d1 = FILTER_SUM_3X3(
		  0,  0,  0,
		  0, -1,  0,
		  0,  0,  1 );
	tjs_int d2 = FILTER_SUM_3X3(
		  0,  0,  0,
		  0,  0,  1,
		  0, -1,  0 );
	if (d1 < 0) d1 = -d1;
	if (d2 < 0) d2 = -d2;
	tjs_int c = (d1 > d2) ? d1 : d2;
	return c;
}



// -------------------------------------------------------------
// G{XtB^
// -------------------------------------------------------------
FILTER_FUNC_DECL(embossFilterR2L3x3)
{
	tjs_int c = EMBOSS_BASE_COLOR_CODE + FILTER_SUM_3X3(
		 -1,  0,  1,
		 -1,  0,  1,
		 -1,  0,  1 );
	DISTORT_COLOR_CODE_MAX(c)
	DISTORT_COLOR_CODE_MIN(c)
	return c;
}

FILTER_FUNC_DECL(embossFilterL2R3x3)
{
	tjs_int c = EMBOSS_BASE_COLOR_CODE + FILTER_SUM_3X3(
		 1,  0, -1,
		 1,  0, -1,
		 1,  0, -1 );
	DISTORT_COLOR_CODE_MAX(c)
	DISTORT_COLOR_CODE_MIN(c)
	return c;
}

FILTER_FUNC_DECL(embossFilterT2B3x3)
{
	tjs_int c = EMBOSS_BASE_COLOR_CODE + FILTER_SUM_3X3(
		 1,  1,  1,
		 0,  0,  0,
		-1, -1, -1 );
	DISTORT_COLOR_CODE_MAX(c)
	DISTORT_COLOR_CODE_MIN(c)
	return c;
}

FILTER_FUNC_DECL(embossFilterB2T3x3)
{
	tjs_int c = EMBOSS_BASE_COLOR_CODE + FILTER_SUM_3X3(
		-1, -1, -1,
		 0,  0,  0,
		 1,  1,  1 );
	DISTORT_COLOR_CODE_MAX(c)
	DISTORT_COLOR_CODE_MIN(c)
	return c;
}

//΂
FILTER_FUNC_DECL(embossFilterLT2RB3x3)
{
	tjs_int c = EMBOSS_BASE_COLOR_CODE + FILTER_SUM_3X3(
		 1,  1,  0,
		 1,  0, -1,
		 0, -1, -1 );
	DISTORT_COLOR_CODE_MAX(c)
	DISTORT_COLOR_CODE_MIN(c)
	return c;
}

FILTER_FUNC_DECL(embossFilterLB2RT3x3)
{
	tjs_int c = EMBOSS_BASE_COLOR_CODE + FILTER_SUM_3X3(
		 0, -1, -1,
		 1,  0, -1,
		 1,  1,  0 );
	DISTORT_COLOR_CODE_MAX(c)
	DISTORT_COLOR_CODE_MIN(c)
	return c;
}

FILTER_FUNC_DECL(embossFilterRT2LB3x3)
{
	tjs_int c = EMBOSS_BASE_COLOR_CODE + FILTER_SUM_3X3(
		 0,  1,  1,
		-1,  0,  1,
		-1, -1,  0 );
	DISTORT_COLOR_CODE_MAX(c)
	DISTORT_COLOR_CODE_MIN(c)
	return c;
}

FILTER_FUNC_DECL(embossFilterRB2LT3x3)
{
	tjs_int c = EMBOSS_BASE_COLOR_CODE + FILTER_SUM_3X3(
		-1, -1,  0,
		-1,  0,  1,
		 0,  1,  1 );
	DISTORT_COLOR_CODE_MAX(c)
	DISTORT_COLOR_CODE_MIN(c)
	return c;
}



// -------------------------------------------------------------
// vVAtB^
// -------------------------------------------------------------
FILTER_FUNC_DECL(laplacianFilter4dir3x3)
{
	tjs_int c = FILTER_SUM_3X3(
		 0, -1,  0,
		-1,  4, -1,
		 0, -1,  0 );
	DISTORT_COLOR_CODE_MAX(c)
	DISTORT_COLOR_CODE_MIN(c)
	return c;
}

FILTER_FUNC_DECL(laplacianFilter8dir3x3)
{
	tjs_int c = FILTER_SUM_3X3(
		-1, -1, -1,
		-1,  8, -1,
		-1, -1, -1 );
	DISTORT_COLOR_CODE_MAX(c)
	DISTORT_COLOR_CODE_MIN(c)
	return c;
}


// -------------------------------------------------------------
// ȂtB^
// -------------------------------------------------------------
FILTER_FUNC_DECL(nonFilter3x3)
{
	tjs_int c = 
		cp2[1];
	return c;
}


// -------------------------------------------------------------
// ͋PxA␳PxJ[R[h悭␳B
static inline tjs_uint32 correctRGB(
	tjs_uint32 col,	// ARGB
	tjs_int yin,	// col̋Px
	tjs_int yout,	// ␳̋Px
	GrayScaleFP_rp grayscale)	// OCXP[֐
{
	tjs_uint32 alpha = (col & 0xff000000);
	tjs_uint32 r = (col & 0x00ff0000) >> 16;
	tjs_uint32 g = (col & 0x0000ff00) >> 8;
	tjs_uint32 b = (col & 0x000000ff) >> 0;

	// RGB␳iPxj
	if (yin > 0) {
		r = (r * yout) / yin;
		g = (g * yout) / yin;
		b = (b * yout) / yin;
	} else {
		r = 0;
		g = 0;
		b = 0;
	}

	// ǂꂩ͂ݏoꍇB
	// ̃[`͖őɋNȂƎvB
	if (r > 255 || g > 255 || b > 255) {
		// ߕ
		tjs_uint32 ro, go, bo;
		ro = go = bo = 0;
		if (r > 255) {
			ro = r - 255;
			r = 255;
		}
		if (g > 255) {
			go = g - 255;
			g = 255;
		}
		if (b > 255) {
			bo = b - 255;
			b = 255;
		}

		// s
		tjs_uint32 rs = 255 - r;
		tjs_uint32 gs = 255 - g;
		tjs_uint32 bs = 255 - b;

		// ePx
		tjs_int yo = grayscale(ro, go, bo);
		tjs_int ys = grayscale(rs, gs, bs);

		if (ys > 0) {
			r += rs * yo / ys;
			g += gs * yo / ys;
			b += bs * yo / ys;
			DISTORT_COLOR_CODE_MAX(r);
			DISTORT_COLOR_CODE_MAX(g);
			DISTORT_COLOR_CODE_MAX(b);
		}
//b=255;r=g=0;
	}
	return ((alpha) | (r << 16) | (g << 8) | (b));
}

// -------------------------------------------------------------



// selector
static GrayScaleFP TPFUNCCALL getGrayScaleFunc(tjs_int gsmode)
{
	GrayScaleFP fp = CopyGrayScaleFactorY709_SD;

#define F__CASE(_cs, _vl) \
	case TP_LUMINANCE_COEFFICIENTS_##_cs: \
		fp = CopyGrayScaleFactor##_vl##_SD; \
		break;

	switch (gsmode) {
		F__CASE(NONE, None)
		F__CASE(Y601, Y601)
		F__CASE(Y709, Y709)
		F__CASE(RGBAVG, RGBAvg)
		default: break;
	}
#undef F__CASE

	return fp;
}


// dǗۂČȂ
// selector
static GrayScaleFP_rp TPFUNCCALL getGrayScaleFuncRP(tjs_int gsmode)
{
	GrayScaleFP_rp fp = GetGrayScaleFactorY709;

#define F__CASE(_cs, _vl) \
	case TP_LUMINANCE_COEFFICIENTS_##_cs: \
		fp = GetGrayScaleFactor##_vl; \
		break;

	switch (gsmode) {
		F__CASE(NONE, None)
		F__CASE(Y601, Y601)
		F__CASE(Y709, Y709)
		F__CASE(RGBAVG, RGBAvg)
		default: break;
	}
#undef F__CASE

	return fp;
}


// selector
static FilterFP TPFUNCCALL getFilterFunc(tjs_int type)
{
	FilterFP fp = nonFilter3x3;

#define F__CASE(_cs, _vl) \
	case TP_FILTER_TYPE_##_cs: \
		fp = _vl; \
		break;

	switch (type) {
		F__CASE(GENERIC, genericFilter3x3)
		F__CASE(GENERIC_AVG, genericAvgFilter3x3)
		F__CASE(PREWITT, prewittFilter3x3)
		F__CASE(SOBEL, sobelFilter3x3)
		F__CASE(ROBERT, robertFilter3x3)
		F__CASE(EMBOSS_L2R, embossFilterL2R3x3)
		F__CASE(EMBOSS_R2L, embossFilterR2L3x3)
		F__CASE(EMBOSS_T2B, embossFilterT2B3x3)
		F__CASE(EMBOSS_B2T, embossFilterB2T3x3)
		F__CASE(EMBOSS_LT2RB, embossFilterLT2RB3x3)
		F__CASE(EMBOSS_LB2RT, embossFilterLB2RT3x3)
		F__CASE(EMBOSS_RT2LB, embossFilterRT2LB3x3)
		F__CASE(EMBOSS_RB2LT, embossFilterRB2LT3x3)
		F__CASE(LAPLACIAN_4DIR, laplacianFilter4dir3x3)
		F__CASE(LAPLACIAN_8DIR, laplacianFilter8dir3x3)
	}
#undef F__CASE

	return fp;
}



/*
OCXP[tB^֐B
3x3̈ɑ΂ătB^{B
ۂ̑Ώۂ3x3̒̃sNẐ݁B

*/
bool TPFUNCCALL Layer_grayFilter3x3Impl(
	LayerInfo * layer,
	tjs_int x,
	tjs_int y,
	tjs_int w,
	tjs_int h,
	tjs_int type,
	tjs_int gsmode,
	tjs_int * kernel)
{
	const tTVPRect * const draw = layer->drawRect;

	PRECHECK_DRAW_AREA(3)

	bool rv = true;

	// J[obt@
	// 
	ALLOC_COLOR_BUFFER_3X3(cbuf, cp, width)

/*
	// HDA?
	ALLOC_COLOR_BUFFER_3X3(abuf, ap, width)
*/

	// OCXP[֐
	const GrayScaleFP grayscale = getGrayScaleFunc(gsmode);
	// tB^֐
	const FilterFP filter = getFilterFunc(type);

	// iAi쐬Ă
	grayscale(layer->getBufferPtr(left, top  ), cp1, width);
	grayscale(layer->getBufferPtr(left, top+1), cp2, width);

	// ϒl΍
	tjs_int ksum = 0;
	for (tjs_int i = 0; i < 9; i++) {
		ksum += kernel[i];
	}
	if (ksum <= 0) ksum = 1;

	// 擪s
	{
		tjs_uint32 * dest = layer->getBufferPtr(left, top);
		for (tjs_int xx = 0; xx < width; xx++)
		{
			tjs_uint32 c = *cp1++;
			GENERATE_COLOR_CODE_FROM_GS(c)
			*dest++ = c;
		}
	}

	for (tjs_int yy = 1; yy < (height-1); yy++)
	{
		tjs_int ymod = yy % 3;
		ASSIGN_COLOR_BUFFER_3X3(cbuf, cp, ymod)

		// i쐬
		grayscale(layer->getBufferPtr(left, top + yy + 1), cp3, width);

		// [
		{
			tjs_uint32 c = *cp2;
			GENERATE_COLOR_CODE_FROM_GS(c)
			LayerBuffer_setPixel_NoCheck(left + 0, top + yy, c, layer);
		}
		// tB^
		for (tjs_int xx = 1; xx < (width-1); xx++)
		{
			tjs_uint32 c = filter(cp1++, cp2++, cp3++, kernel, ksum);
			GENERATE_COLOR_CODE_FROM_GS(c)
			LayerBuffer_setPixel_NoCheck(left + xx, top + yy, c, layer);
		}
		// E[
		{
			tjs_uint32 c = *cp2;
			GENERATE_COLOR_CODE_FROM_GS(c)
			LayerBuffer_setPixel_NoCheck(left + width - 1, top + yy, c, layer);
		}
	}

	// ŏIs
	{
		tjs_uint32 * dest = layer->getBufferPtr(left, top + height - 1);
		// E[O܂Őił|C^[܂Ŗ߂B
		// ȏB
		cp3 -= (width - 2);
		for (tjs_int xx = 0; xx < width; xx++)
		{
			tjs_uint32 c = *cp3++;
			GENERATE_COLOR_CODE_FROM_GS(c)
			*dest++ = c;
		}
	}

	// 
	FREE_COLOR_BUFFER_3X3(cbuf)

/*
	// HDA?
	FREE_COLOR_BUFFER_3X3(abuf)
*/

	// auto update
	if (TPFAutoUpdateMode && rv) {
		Layer_FuncCall_update(layer, draw->left, draw->top, draw->get_width(), draw->get_height());
	}

	return rv;
}




// CRC[[V
bool TPFUNCCALL Layer_equalizeImpl(
	LayerInfo * layer,
	tjs_int x,
	tjs_int y,
	tjs_int w,
	tjs_int h,
	tjs_int gsmode,
	bool withColor)	// J[[hŃCRC[[V
{

	const tTVPRect * const draw = layer->drawRect;

	PRECHECK_DRAW_AREA(1)

	bool rv = false;

	// OCXP[֐
	const GrayScaleFP grayscale = getGrayScaleFunc(gsmode);
	const tjs_int numPixels = width * height;
	const tjs_int numColors = 256;

	tjs_int * factor = new tjs_int[numPixels];	// OCXP[p^[
	tjs_uint32 * freq   = new tjs_uint32[numColors];	// J[R[hgpx

	// OCXP[ϊ
	for (tjs_int yy = 0; yy < height; yy++) {
		grayscale(layer->getBufferPtr(left, top + yy), &(factor[width * yy]), width);
	}

	// 0ŏ
	for (tjs_int i = 0; i < numColors; i++) {
		freq[i] = 0;
	}

	// xZo
	for (tjs_int i = 0; i < numPixels; i++) {
		freq[factor[i]]++;
	}

	// ݐϓxɕϊ
	for (tjs_int i = 1; i < numColors; i++) {
		freq[i] += freq[i-1];
	}

	// ŏ̗ݐϓx
	tjs_uint32 nmin = freq[0];

	// fƍŏݐϓx̍
	tjs_uint32 nfac = numPixels - nmin;

	if (nfac > 0 ) {
		if (withColor) {
			// OCXP[֐
			const GrayScaleFP_rp grayscale_rp = getGrayScaleFuncRP(gsmode);
			for (tjs_int yy = 0; yy < height; yy++) {
				tjs_uint32 * p = layer->getBufferPtr(left, top + yy);
				tjs_int idx = yy * width;
				for (tjs_int xx = 0; xx < width; xx++) {
					tjs_int c0 = factor[idx + xx];
					tjs_int c = (numColors - 1) * (freq[c0] - nmin) / nfac;
					p[xx] = correctRGB(p[xx], c0, c, grayscale_rp);
				}
			}
		} else {
			for (tjs_int yy = 0; yy < height; yy++) {
				tjs_uint32 * p = layer->getBufferPtr(left, top + yy);
				tjs_int idx = yy * width;
				for (tjs_int xx = 0; xx < width; xx++) {
					tjs_uint32 c = (numColors - 1) * (freq[factor[idx + xx]] - nmin) / nfac;
					GENERATE_COLOR_CODE_FROM_GS(c)
					p[xx] = c;
				}
			}
		}
		rv = true;
	}

	delete[] factor;
	delete[] freq;

	// auto update
	if (TPFAutoUpdateMode && rv) {
		Layer_FuncCall_update(layer, draw->left, draw->top, draw->get_width(), draw->get_height());
	}

	return rv;
}




// ėpTCỸUCNtB^BȂ񂾂dȁ[B
static void TPFUNCCALL genericMosaicFilter(
	tjs_int sx,	// Jn_
	tjs_int sy,	// Jn_
	tjs_int width,	// UCNTCY
	tjs_int height,	// UCNTCY
	LayerInfo * layer)
{
	tjs_uint32 r = 0, g = 0, b = 0;

	// RGBeƂɘaZo
	for (tjs_int y = 0; y < height; y++) {
		tjs_uint32 * p = layer->getBufferPtr(sx, sy + y);
		for (tjs_int x = 0; x < width; x++) {
			tjs_uint32 c = *p++;
			r += (c & 0x00ff0000) >> 16;
			g += (c & 0x0000ff00) >> 8;
			b += (c & 0x000000ff) >> 0;
		}
	}

	// RGBϒl
	{
		tjs_uint32 sz = width * height;
		r /= sz;
		g /= sz;
		b /= sz;
	}

	// ̈hԂ
	tjs_uint32 v = (0xff000000) | (r << 16) | (g << 8) | (b);
	for (tjs_int y = 0; y < height; y++) {
		tjs_uint32 * p = layer->getBufferPtr(sx, sy + y);
		for (tjs_int x = 0; x < width; x++) {
			*p++ = v;
		}
	}

	return;
}


// UCNtB^
static bool TPFUNCCALL Layer_colorFilterMosaicImpl(
	LayerInfo * layer,
	tjs_int x,
	tjs_int y,
	tjs_int w,
	tjs_int h,
	tjs_int size)
{
	const tTVPRect * const draw = layer->drawRect;

	PRECHECK_DRAW_AREA(1)

	bool rv = true;

	const tjs_int quotientX  = width  / size;
	const tjs_int quotientY  = height / size;
	const tjs_int remainderX = width  % size;
	const tjs_int remainderY = height % size;

	// ʏs
	for (tjs_int qy = 0; qy < quotientY; qy++)
	{
		tjs_int yy = top + qy * size;
		tjs_int xx = left;

		// tB^
		for (tjs_int qx = 0; qx < quotientX; qx++)
		{
			genericMosaicFilter(xx, yy, size, size, layer);
			xx += size;
		}
		// E[
		if (remainderX) {
			genericMosaicFilter(xx, yy, remainderX, size, layer);
		}
	}

	// ŉs
	if (remainderY) {
		tjs_int yy = top + quotientY * size;
		tjs_int xx = left;

		// tB^
		for (tjs_int qx = 0; qx < quotientX; qx++)
		{
			genericMosaicFilter(xx, yy, size, remainderY, layer);
			xx += size;
		}
		// E[
		if (remainderX) {
			genericMosaicFilter(xx, yy, remainderX, remainderY, layer);
		}
	}

	// auto update
	if (TPFAutoUpdateMode && rv) {
		Layer_FuncCall_update(layer, draw->left, draw->top, draw->get_width(), draw->get_height());
	}

	return rv;
}





// NCbN\[g
template <typename T>
static void TPFASTCALL quicksort(T * a, tjs_int left, tjs_int right)
{
	if (left < right) {
		tjs_int somewhere = (left + right) / 2;
		T pivot = a[somewhere];
		a[somewhere] = a[left];
		tjs_int p = left;
		for (tjs_int i = left + 1; i <= right; i++) {
			if (a[i] < pivot) {
				p++;
				T tmp = a[p];
				a[p] = a[i];
				a[i] = tmp;
			}
		}
		a[left] = a[p];
		a[p] = pivot;
		quicksort<T>(a, left, p-1);
		quicksort<T>(a, p+1, right);
	}
}


/*
3x3̈ɑ΂ăfBAtB^{B
ۂ̑Ώۂ3x3̒̃sNẐ݁B
O1sNZ͏ȂB
*/
bool TPFUNCCALL Layer_colorFilterMedian3x3Impl(
	LayerInfo * layer,
	tjs_int x,
	tjs_int y,
	tjs_int w,
	tjs_int h)
{
	const tTVPRect * const draw = layer->drawRect;

	PRECHECK_DRAW_AREA(3)

	bool rv = true;

	// J[obt@
	// 
	ALLOC_COLOR_BUFFER_3X3(rbuf, rp, width)
	ALLOC_COLOR_BUFFER_3X3(gbuf, gp, width)
	ALLOC_COLOR_BUFFER_3X3(bbuf, bp, width)

	// iAi쐬Ă
	{
		tjs_uint32 * p;
		p = layer->getBufferPtr(left, top);
		COPY_GRAYSCALE_FACTOR_RGB(p, rp1, gp1, bp1, width)
		p = layer->getBufferPtr(left, top+1);
		COPY_GRAYSCALE_FACTOR_RGB(p, rp2, gp2, bp2, width)
	}

	for (tjs_int yy = 1; yy < (height-1); yy++)
	{
		{
			tjs_int ymod = yy % 3;
			ASSIGN_COLOR_BUFFER_3X3(rbuf, rp, ymod)
			ASSIGN_COLOR_BUFFER_3X3(gbuf, gp, ymod)
			ASSIGN_COLOR_BUFFER_3X3(bbuf, bp, ymod)
		}

		// i쐬
		{
			tjs_uint32 * p = layer->getBufferPtr(left, top + yy + 1);
			COPY_GRAYSCALE_FACTOR_RGB(p, rp3, gp3, bp3, width)
		}

		// tB^

#define SET_MEDIAN_3X3(_qsp, _xp) \
			tjs_uint32 _qsp[9] = { _xp##1[0], _xp##1[1], _xp##1[2], _xp##2[0], _xp##2[1], _xp##2[2], _xp##3[0], _xp##3[1], _xp##3[2] }; \
			_xp##1++; \
			_xp##2++; \
			_xp##3++; \
			quicksort<tjs_uint32>(_qsp, 0, 8);

		for (tjs_int xx = 1; xx < (width-1); xx++)
		{
			SET_MEDIAN_3X3(r_qs, rp)
			SET_MEDIAN_3X3(g_qs, gp)
			SET_MEDIAN_3X3(b_qs, bp)

			tjs_uint32 c = 0xff000000 | (r_qs[4] << 16) | (g_qs[4] << 8) | (b_qs[4]);
			LayerBuffer_setPixel_NoCheck(left + xx, top + yy, c, layer);
		}

#undef SET_MEDIAN_3X3

	}

	// 
	FREE_COLOR_BUFFER_3X3(rbuf)
	FREE_COLOR_BUFFER_3X3(gbuf)
	FREE_COLOR_BUFFER_3X3(bbuf)

	// auto update
	if (TPFAutoUpdateMode && rv) {
		Layer_FuncCall_update(layer, draw->left, draw->top, draw->get_width(), draw->get_height());
	}

	return rv;
}




// ėpJ[tB^
// 
bool TPFUNCCALL Layer_colorFilterImpl(
	LayerInfo * layer,
	tjs_int x,
	tjs_int y,
	tjs_int w,
	tjs_int h,
	tjs_int type,
	tjs_int size,
	tjs_int * kernel)
{
	bool rv = false;
	switch (type) {
	case TP_FILTER_TYPE_MOSAIC:
		rv = Layer_colorFilterMosaicImpl(layer, x, y, w, h, size);
		break;
	case TP_FILTER_TYPE_MEDIAN:
		rv = Layer_colorFilterMedian3x3Impl(layer, x, y, w, h);
		break;
	default:
		break;
	}

	return rv;
}



#undef EMBOSS_BASE_COLOR_CODE
#undef FILTER_SUM_3X3
#undef DISTORT_COLOR_CODE_MAX
#undef DISTORT_COLOR_CODE_MIN
#undef GENERATE_COLOR_CODE_FROM_GS
#undef FILTER_FUNC_DECL
#undef PRECHECK_DRAW_AREA
#undef ASSIGN_COLOR_BUFFER_3X3
#undef ALLOC_COLOR_BUFFER_3X3
#undef FREE_COLOR_BUFFER_3X3
