#!perl
# tvpgl.* generator

# 関数ジェネレータ

use utf8;
my $ENCODE_NAME = ':encoding(shift_jis)';
binmode(STDIN,  $ENCODE_NAME);
binmode(STDOUT, $ENCODE_NAME);

my $Csrc = '';

my $FileNameMT = '../common/maketab.c';
my $FileNameC = 'TPGraphic.cpp';
my $FileNameH = 'TPGraphic.h';
my $HeaderID = '_TVP_PLUGIN_TP_GRAPHIC_H_';

$gpl = <<EOF;

EOF

$rev = "0.1";

;#-----------------------------------------------------------------
;# file content retrieving
;#-----------------------------------------------------------------
sub get_file_content
{
	local($name, @all);
	$name = $_[0];
	open IFH, $name;
	binmode(IFH, $ENCODE_NAME);
	@all = <IFH>;
	close IFH;
	return join('', @all);
}

;#-----------------------------------------------------------------
;# exported function gathering
;#-----------------------------------------------------------------


sub format_function
{
	local($a, $b);
	$a = $_[0];
	$a =~ s/\n//g;
	push (@function_list, $a);
}

sub get_function_list
{
	@function_list = ();
	local($content, $ret);
	$content = &get_file_content($_[0]);
	$ret = '';
	$content =~ s/\/\*export\*\/(.*?)\{/&format_function($1)/seg;
}

;#-----------------------------------------------------------------
;# loop unrolling
;#-----------------------------------------------------------------

sub loop_unroll_c
{
	local($content, $times, $unroll_times, $i);
	$content = $_[0];
	$times = $_[1];
	$unroll_times = $_[2];

	$Csrc .= <<EOF;
  if($times > 0)
  {
	int lu_n = ($times + ($unroll_times-1)) / $unroll_times;
	switch($times % $unroll_times)
	{
	case 0: do { $content;
EOF
	for($i = $unroll_times-1; $i>=1; $i--)
	{
		$Csrc .= <<EOF;
	case $i: $content;
EOF
	}
	$Csrc .= <<EOF;
	   } while(-- lu_n);
	}
  }
EOF
}

;# type 2 for more simple case
;# backward operation
sub loop_unroll_c_2_backward
{
	local($content, $times, $unroll_times, $i, $con);
	$content = $_[0];
	$times = $_[1];
	$unroll_times = $_[2];

	$Csrc .= <<EOF;
	$times --;

	while($times >= ($unroll_times -1))
	{
EOF

	for($i = 0; $i<$unroll_times; $i++)
	{
		$con = $content;
		$con =~ s/\{ofs\}/($times-$i)/g;
		$Csrc .= $con;
	}

  	$Csrc .= <<EOF;
		$times -= $unroll_times;
	}

	while($times >= 0)
	{
EOF

	$con = $content;
	$con =~ s/\{ofs\}/$times/g;
	$Csrc .= $con;

  	$Csrc .= <<EOF;
		$times --;
	}
EOF
}

;# type 2 forward
sub loop_unroll_c_2
{
	local($content, $times, $unroll_times, $i, $con);
	$content = $_[0];
	$times = $_[1];
	$unroll_times = $_[2];

	$Csrc .= <<EOF;
	{
		int ___index = 0;
		$times -= ($unroll_times-1);

		while(___index < $times)
		{
EOF

	for($i = 0; $i<$unroll_times; $i++)
	{
		$con = $content;
		$con =~ s/\{ofs\}/(___index+$i)/g;
		$Csrc .= $con;
	}

  	$Csrc .= <<EOF;
			___index += $unroll_times;
		}

		$times += ($unroll_times-1);

		while(___index < $times)
		{
EOF

		$con = $content;
		$con =~ s/\{ofs\}/___index/g;
		$Csrc .= $con;

  		$Csrc .= <<EOF;
			___index ++;
		}
	}
EOF
}


;# twice-interleaved type of type 2
sub loop_unroll_c_int_2
{
	local($content, $times, $content2, $int_times, $unroll_times, $i, $con, $con2, @cont, @cont2);
	$content = $_[0];
	$content2 = $_[1];
	$times = $_[2];
	$unroll_times = $_[3];

	$Csrc .= <<EOF;
	{
		int ___index = 0;
		$times -= ($unroll_times-1);

		while(___index < $times)
		{
EOF

	for($i = 0; $i<$unroll_times/2; $i++)
	{
		local($j);
		$con = $content;
		$con =~ s/\{ofs\}/(___index+($i*2))/g;
		$con2 = $content2;
		$con2 =~ s/\{ofs\}/(___index+($i*2+1))/g;
		$con =~ s/\n//g;
		$con2 =~ s/\n//g;
		@cont = split(/;;/, $con);
		@cont2 = split(/;;/, $con2);

		for($j = 0; $j <= $#cont; $j++)
		{
			$Csrc .= $cont[$j];
			$Csrc .= ";\n";
			$Csrc .= $cont2[$j];
			$Csrc .= ";\n";
		}
	}

  	$Csrc .= <<EOF;
			___index += $unroll_times;
		}

		$times += ($unroll_times-1);

		while(___index < $times)
		{
EOF

	$con = $content;
	$con =~ s/\{ofs\}/___index/g;
	$Csrc .= $con;

  	$Csrc .= <<EOF;
			___index ++;
		}
	}
EOF
}



$serial = 0;

sub _label_expand
{
	local($num);
	$num = $_[0];
	return "ll_". ($num + $serial);
}

sub label_expand
{
	local($content);
	$content = $_[0];
	$serial += ($content =~ s/\{(\d+)\}/&_label_expand($1)/eg);
	return $content;
}

sub _adds_expand
{
	local($reg, $num, $mul);
	$reg = $_[0];
	$num = $_[1];
	$mul = $_[2];
	$num *= $mul;
	if($num == -1)
	{
		return "dec $reg";
	}
	elsif($num == 1)
	{
		return "inc $reg";
	}
	else
	{
		return "add $reg, $num";
	}
}

sub adds_expand
{
	local($content, $mul);
	$content = $_[0];
	$mul = $_[1];
	$content =~ s/\{(.+?),(\d+)\}/&_adds_expand($1, $2, $mul)/eg;
	return $content;
}






;#-----------------------------------------------------------------
;# grayscale conversion
;#-----------------------------------------------------------------

# --------------------------
# Y709 HDTV
# --------------------------
$Csrc .= << 'EOF';
// Src-Dest Style
/*TP*/GRAYSCALE_FUNC_DECL(CopyGrayScaleFactorY709_SD)
{
	tjs_uint32 s1, d1, s2, d2;
EOF

$content = << 'EOF';
	s1 = src[{ofs}];;
	d1 = (s1&0xff)*19;;
	d1 += ((s1 >> 8)&0xff)*183;;
	d1 += ((s1 >> 16)&0xff)*54;;
	d1 >>= 8;;
	dest[{ofs}] = d1;;
EOF
$content2 = << 'EOF';
	s2 = src[{ofs}];;
	d2 = (s2&0xff)*19;;
	d2 += ((s2 >> 8)&0xff)*183;;
	d2 += ((s2 >> 16)&0xff)*54;;
	d2 >>= 8;;
	dest[{ofs}] = d2;;
EOF

&loop_unroll_c_int_2($content, $content2, 'len', 4);

$Csrc .= << 'EOF';
}

EOF

$Csrc .= << 'EOF';
/*TP*/GRAYSCALE_RP_FUNC_DECL(GetGrayScaleFactorY709)
{
	return ((r*19 + g*183 + b*54) >> 8);
}
EOF




# --------------------------
# Y601 NTSC
# --------------------------
$Csrc .= << 'EOF';
// Y601 Src-Dest Style
/*TP*/GRAYSCALE_FUNC_DECL(CopyGrayScaleFactorY601_SD)
{
	tjs_uint32 s1, d1, s2, d2;
EOF

$content = << 'EOF';
	s1 = src[{ofs}];;
	d1 = (s1&0xff)*29;;
	d1 += ((s1 >> 8)&0xff)*150;;
	d1 += ((s1 >> 16)&0xff)*77;;
	d1 >>= 8;;
	dest[{ofs}] = d1;;
EOF
$content2 = << 'EOF';
	s2 = src[{ofs}];;
	d2 = (s2&0xff)*29;;
	d2 += ((s2 >> 8)&0xff)*150;;
	d2 += ((s2 >> 16)&0xff)*77;;
	d2 >>= 8;;
	dest[{ofs}] = d2;;
EOF

&loop_unroll_c_int_2($content, $content2, 'len', 4);

$Csrc .= << 'EOF';
}

EOF

$Csrc .= << 'EOF';
/*TP*/GRAYSCALE_RP_FUNC_DECL(GetGrayScaleFactorY601)
{
	return ((r*29 + g*150 + b*77) >> 8);
}
EOF



# --------------------------
# Average RGB
# --------------------------
$Csrc .= << 'EOF';
//  Src-Dest Style
/*TP*/GRAYSCALE_FUNC_DECL(CopyGrayScaleFactorRGBAvg_SD)
{
	tjs_uint32 s1, d1, s2, d2;
EOF

$content = << 'EOF';
	s1 = src[{ofs}];;
	d1 = (s1&0xff);;
	d1 += ((s1 >> 8)&0xff);;
	d1 += ((s1 >> 16)&0xff);;
	d1 /= 3;;
	dest[{ofs}] = d1;;
EOF
$content2 = << 'EOF';
	s2 = src[{ofs}];;
	d2 = (s2&0xff);;
	d2 += ((s2 >> 8)&0xff);;
	d2 += ((s2 >> 16)&0xff);;
	d2 /= 3;;
	dest[{ofs}] = d2;;
EOF

&loop_unroll_c_int_2($content, $content2, 'len', 4);

$Csrc .= << 'EOF';
}

EOF


$Csrc .= << 'EOF';
/*TP*/GRAYSCALE_RP_FUNC_DECL(GetGrayScaleFactorRGBAvg)
{
	return ((r + g + b)/3);
}
EOF



# --------------------------
# None: 
# --------------------------
$Csrc .= << 'EOF';
// None
/*TP*/GRAYSCALE_FUNC_DECL(CopyGrayScaleFactorNone_SD)
{
	tjs_uint32 d1, d2;
EOF

$content = << 'EOF';
	d1 = src[{ofs}]&0xff;;
	dest[{ofs}] = d1;;
EOF
$content2 = << 'EOF';
	d2 = src[{ofs}]&0xff;;
	dest[{ofs}] = d2;;
EOF

&loop_unroll_c_int_2($content, $content2, 'len', 4);

$Csrc .= << 'EOF';
}

EOF

$Csrc .= << 'EOF';
/*TP*/GRAYSCALE_RP_FUNC_DECL(GetGrayScaleFactorNone)
{
	return b;
}
EOF


# --------------------------
# Only Red: 
# --------------------------
$Csrc .= << 'EOF';
/*TP*/GRAYSCALE_FUNC_DECL(CopyGrayScaleFactorOnlyRed_SD)
{
	tjs_uint32 d1, d2;
EOF

$content = << 'EOF';
	d1 = src[{ofs}]&0x00ff0000;;
	dest[{ofs}] = d1 >> 16;;
EOF
$content2 = << 'EOF';
	d2 = src[{ofs}]&0x00ff0000;;
	dest[{ofs}] = d2 >> 16;;
EOF

&loop_unroll_c_int_2($content, $content2, 'len', 4);

$Csrc .= << 'EOF';
}

EOF

# --------------------------
# Only Green: 
# --------------------------
$Csrc .= << 'EOF';
/*TP*/GRAYSCALE_FUNC_DECL(CopyGrayScaleFactorOnlyGreen_SD)
{
	tjs_uint32 d1, d2;
EOF

$content = << 'EOF';
	d1 = src[{ofs}]&0x0000ff00;;
	dest[{ofs}] = d1 >> 8;;
EOF
$content2 = << 'EOF';
	d2 = src[{ofs}]&0x0000ff00;;
	dest[{ofs}] = d2 >> 8;;
EOF

&loop_unroll_c_int_2($content, $content2, 'len', 4);

$Csrc .= << 'EOF';
}

EOF


# --------------------------
# Only Blue: 
# --------------------------
$Csrc .= << 'EOF';
/*TP*/GRAYSCALE_FUNC_DECL(CopyGrayScaleFactorOnlyBlue_SD)
{
	tjs_uint32 d1, d2;
EOF

$content = << 'EOF';
	d1 = src[{ofs}]&0x000000ff;;
	dest[{ofs}] = d1;;
EOF
$content2 = << 'EOF';
	d2 = src[{ofs}]&0x000000ff;;
	dest[{ofs}] = d2;;
EOF

&loop_unroll_c_int_2($content, $content2, 'len', 4);

$Csrc .= << 'EOF';
}

EOF




# --------------------------
# Linear Transformation
# --------------------------
# dest: destination address
# len: width (>0)
# src: source image buffer pointer
# sx, sy: coords on source image buffer (1bit+-15bit.16bit signed fixed point number)
# stepx, stepy: step width, step height (1bit+-15bit.16bit signed fixed point number)
# srcpitch: image buffer pitch in byte (>0)
# opa: opacity (0-255)
$Csrc .= << 'EOF';
/*TP*/void TVPPlugInLinTransConstAlphaBlend_SD(tjs_uint32 *dest, tjs_int len, const tjs_uint32 *src1, tjs_int sx1, tjs_int sy1, tjs_int stepx1, tjs_int stepy1, tjs_int srcpitch1, const tjs_uint32 *src2, tjs_int sx2, tjs_int sy2, tjs_int stepx2, tjs_int stepy2, tjs_int srcpitch2, tjs_int opa)
{
	tjs_uint32 s1, s2, d1;
EOF

$content = <<EOF;
	{
	s1 = *( (const tjs_uint32*)((const tjs_uint8*)src1 + (sy1>>16)*srcpitch1) + (sx1>>16));
	s2 = *( (const tjs_uint32*)((const tjs_uint8*)src2 + (sy2>>16)*srcpitch2) + (sx2>>16));
	sx1 += stepx1;
	sy1 += stepy1;
	sx2 += stepx2;
	sy2 += stepy2;
	d1 = s1 & 0xff00ff;
	d1 = (d1 + (((s2 & 0xff00ff) - d1) * opa >> 8)) & 0xff00ff;
	s1 &= 0xff00;
	s2 &= 0xff00;
	*dest = d1 | ((s1 + ((s2 - s1) * opa >> 8)) & 0xff00);
	dest++;
	}
EOF

&loop_unroll_c($content, 'len', 4);

$Csrc .= << 'EOF';
}

EOF


;#-----------------------------------------------------------------
# RGB/HSV関連


# RGB成分の最大値を求める
$Csrc .= << 'EOF';

/*TP*/inline tjs_int TVPPlugInGetMaxRGB(tjs_uint32 col)
{
	tjs_int max = (col & 0xFF0000) >> 16;
	tjs_int g   = (col & 0x00FF00) >> 8;
	tjs_int b   = (col & 0x0000FF);
	if (g > max) max = g;
	if (b > max) max = b;
	return max;
}

EOF

# RGB成分の最大値を求める
$Csrc .= << 'EOF';

/*TP*/inline tjs_int TVPPlugInGetMaxRGB(tjs_int r, tjs_int g, tjs_int b)
{
	tjs_int max = r;
	if (g > max) max = g;
	if (b > max) max = b;
	return max;
}

EOF

# RGB成分の最小値を求める
$Csrc .= << 'EOF';

/*TP*/inline tjs_int TVPPlugInGetMinRGB(tjs_uint32 col)
{
	tjs_int min = (col & 0xFF0000) >> 16;
	tjs_int g   = (col & 0x00FF00) >> 8;
	tjs_int b   = (col & 0x0000FF);
	if (g < min) min = g;
	if (b < min) min = b;
	return min;
}

EOF

# RGB成分の最小値を求める
$Csrc .= << 'EOF';

/*TP*/inline tjs_int TVPPlugInGetMinRGB(tjs_int r, tjs_int g, tjs_int b)
{
	tjs_int min = r;
	if (g < min) min = g;
	if (b < min) min = b;
	return min;
}

EOF

# RGB成分の最大値と最小値を求める
$Csrc .= << 'EOF';

/*TP*/inline void TVPPlugInGetMinMaxRGB(tjs_int r, tjs_int g, tjs_int b, tjs_int & min, tjs_int & max)
{
	min = max = r;
	if (g > max) max = g;
	if (b > max) max = b;
	if (g < min) min = g;
	if (b < min) min = b;
	return;
}

EOF

# Taken from "Fund'l of 3D Computer Graphics", Alan Watt (1989)
#   Assumes (r,g,b) range from 0 to 255
#   Sets h in degrees: [0,360]
#        s in [0,255]
#        v in [0,255]
$Csrc .= << 'EOF';

/*TP*/inline void TVPPlugInConvertRGB2HSV(tjs_int r, tjs_int g, tjs_int b, tjs_int & hout, tjs_int & sout, tjs_int & vout)
{
	const tjs_int max_v = TVPPlugInGetMaxRGB(r, g, b);
	const tjs_int min_v = TVPPlugInGetMinRGB(r, g, b);
	const tjs_int diff = max_v - min_v;

	// Lightness
	tjs_int v = max_v;

	// Saturation
	tjs_int s = 0;
	if (max_v > 0) {
		s = 255 * diff / max_v;
	} else {
		s = 0;
	}

	// Hue
	tjs_int h = 0;
	if (s > 0) {
		// diff > 0
		if (r == max_v) {
			h = 60 * (0        + g - b) / diff;
			if (h < 0) h += 360;
		} else if (g == max_v) {
			h = 60 * (2 * diff + b - r) / diff;
		} else if (b == max_v) {
			h = 60 * (4 * diff + r - g) / diff;
		}
	}

	hout = h;
	sout = s;
	vout = v;

	return;
}

EOF

$Csrc .= << 'EOF';

/*TP*/inline void TVPPlugInConvertRGB2HSV(tjs_uint32 col, tjs_int & hout, tjs_int & sout, tjs_int & vout)
{
	TVPPlugInConvertRGB2HSV((col & 0xFF0000) >> 16, (col & 0x00FF00) >> 8, (col & 0x0000FF), hout, sout, vout);
}

EOF




#-----------------------------------------------------------------
# ARGBコピー
#-----------------------------------------------------------------

$Csrc .= << 'EOF';

/*TP*/void TVPPlugInCopyARGB(tjs_uint32 *dest, const tjs_uint32 *src, tjs_int len)
{
EOF

$content = <<EOF;
	dest[{ofs}] = src[{ofs}];;
EOF

&loop_unroll_c_2($content, 'len', 8);

$Csrc .= << 'EOF';
}

EOF


;#-----------------------------------------------------------------

# 1ピクセル専用関数
# naming rule: TVPPlugInExportFunc_1px()

$Csrc .= << 'EOF';

/*TP*/inline tjs_uint32 TVPPlugInMulColor(tjs_uint32 color, tjs_uint32 fac)
{
	return (((((color & 0x00ff00) * fac) & 0x00ff0000) +
			(((color & 0xff00ff) * fac) & 0xff00ff00) ) >> 8);
}

/*TP*/inline tjs_uint32 TVPPlugInSaturatedAdd(tjs_uint32 a, tjs_uint32 b)
{
	/* Add each byte of packed 8bit values in two 32bit uint32, with saturation. */
	tjs_uint32 tmp = (  ( a & b ) + ( ((a ^ b)>>1) & 0x7f7f7f7f)  ) & 0x80808080;
	tmp = (tmp<<1) - (tmp>>7);
	return (a + b - tmp) | tmp;
}

/*TP*/inline tjs_uint32 TVPPlugInAddAlphaBlend_a_ca(tjs_uint32 dest, tjs_uint32 sopa, tjs_uint32 sopa_inv, tjs_uint32 src)
{
	/*
		Di = sat(Si, (1-Sa)*Di)
		Da = Sa + Da - SaDa
	*/

	tjs_uint32 dopa = dest >> 24;
	dopa = dopa + sopa - (dopa*sopa >> 8);
	dopa -= (dopa >> 8); /* adjust alpha */
	return (dopa << 24) + 
		TVPPlugInSaturatedAdd((((dest & 0xff00ff)*sopa_inv >> 8) & 0xff00ff) +
			(((dest & 0xff00)*sopa_inv >> 8) & 0xff00), src);
}


// normal alpha
// table: TVPOpacityOnOpacityTable
/*TP*/void TVPPlugInConstColorAlphaBlend_d_1px(tjs_uint32 *dest, tjs_uint32 color, tjs_int opa, const unsigned char * table)
{
	tjs_uint32 d1, s1, d, dopa;
	tjs_int alpha;
	s1 = color & 0xff00ff;
	color = color & 0xff00;

	d = *dest;
	dopa = d>>24;
	alpha = table[dopa + (opa<<8)];
	d1 = d & 0xff00ff;
	d1 = ((d1 + ((s1 - d1) * alpha >> 8)) & 0xff00ff) |
		((255-((255-dopa)*(255-opa)>>8)) << 24);
	d &= 0xff00;
	*dest = d1 | ((d + ((color - d) * alpha >> 8)) & 0xff00);
	return;
}

// addalpha
/*TP*/inline void TVPPlugInConstColorAlphaBlend_a_1px(tjs_uint32 *dest, tjs_uint32 color, tjs_int opa)
{
	tjs_uint32 src = TVPPlugInMulColor(color & 0xffffff, opa);
	tjs_uint32 opa_inv = opa ^ 0xff;

	*dest = TVPPlugInAddAlphaBlend_a_ca(*dest, opa, opa_inv, src);
	return;
}


/*TP*/inline void TVPPlugInConstColorAlphaBlend_1px(tjs_uint32 *dest, tjs_uint32 color, tjs_int opa)
{
	/* this function always holds desitination alpha channel */
	tjs_uint32 s1 = (color & 0xff00ff)*opa ;
	tjs_uint32 d = *dest;

	color = (color & 0xff00)*opa ;
	opa = 255 - opa;

	*dest = (d & 0xff000000) + ((((d & 0xff00ff) * opa + s1) >> 8) & 0xff00ff) +
		((((d&0xff00) * opa + color) >> 8) & 0xff00);
}



/*TP*/inline void TVPPlugInFillARGB_1px(tjs_uint32 *dest, tjs_uint32 value)
{
	*dest = value;
}


/*TP*/inline void TVPPlugInFillColor_1px(tjs_uint32 *dest, tjs_uint32 color)
{
	color &= 0xffffff;
	tjs_uint32 t1 = *dest;
	t1 &= 0xff000000;
	t1 += color;
	*dest = t1;
}


/*TP*/inline void TVPPlugInFillMask_1px(tjs_uint32 *dest, tjs_uint32 mask)
{
	mask <<= 24;
	tjs_uint32 t1 = *dest;
	t1 &= 0x00ffffff;
	t1 += mask;
	*dest = t1;
}

EOF

;#-----------------------------------------------------------------

# write to FC
{
	# inlineをのぞく
	my @S = split(/\n/, $Csrc);
	my @D = ();
	while (@S) {
		my $s = shift @S;
		if ($s =~ m|^/\*TP\*/inline|) {
			while (@S) {
				my $s = shift @S;
				last if ($s eq '}');
			}
		} else {
			push @D, $s;
		}
	}

	open FC, ">$FileNameC";
	binmode(FC, $ENCODE_NAME);

	print FC << "EOF";
//----------------------------
// Generated by $0
//----------------------------

/*
$gpl
*/

#include "$FileNameH"

EOF

	my $tab = get_file_content($FileNameMT);
	# staticをはずす
	$tab =~ s/static\s+void\s+TVPCreateTable/void TVPCreateTable/;
	# 関数名を変更する
	$tab =~ s/TVP([_a-zA-Z0-9]+)\(/TVPPlugIn$1(/g;

	print FC << "END";
//----------------------------
// $FileNameMT パクり ここから
//----------------------------
$tab
//----------------------------
// $FileNameMT パクり ここまで
//----------------------------
END

	print FC join("\n", @D);
	close FC;
}

;#-----------------------------------------------------------------

open FH, ">$FileNameH";
binmode(FH, $ENCODE_NAME);

print FH << "EOF";
#ifndef $HeaderID
#define $HeaderID

#include "TVPPlugIn.h"

// 誰か一回だけ呼んでくださいお願いします
void TVPPlugInCreateTable(void);
// オパオパ
extern unsigned char TVPOpacityOnOpacityTable[256*256];


#define  TP_LUMINANCE_COEFFICIENTS_NONE    0
#define  TP_LUMINANCE_COEFFICIENTS_Y601  601
#define  TP_LUMINANCE_COEFFICIENTS_Y709  709
#define  TP_LUMINANCE_COEFFICIENTS_RGBAVG  1

#define GRAYSCALE_FUNC_DECL(_func) void _func(tjs_uint32 *src, tjs_int *dest, tjs_int len)
#define GRAYSCALE_RP_FUNC_DECL(_func) tjs_int _func(tjs_uint32 r, tjs_uint32 g, tjs_uint32 b)

typedef void (*GrayScaleFP)(tjs_uint32 *, tjs_int *, tjs_int);
typedef tjs_int (*GrayScaleFP_rp)(tjs_uint32, tjs_uint32, tjs_uint32);

EOF

{
	# inlineを取得
	my @S = split(/\n/, $Csrc);
	my @D = ();
	while (@S) {
		my $s = shift @S;
		if ($s =~ m|^/\*TP\*/inline|) {
			push @D, $s;
			while (@S) {
				my $s = shift @S;
				push @D, $s;
				last if ($s eq '}');
			}
		} elsif ($s =~ m|^/\*TP\*/|) {
			push @D, "$s;";
		}
	}

	print FH join("\n", @D);
}


print FH << "EOF";

#endif  $HeaderID
EOF

close FH;
;#-----------------------------------------------------------------
