#include "SQLite3Intf.h"


// class SQLite3ConnectionIntf

// open DB
void SQLite3ConnectionIntf::open(tTJSString & file)
{
	if (db == NULL) {
		db = new sqlite3plus::SqliteConnection(file.c_str());
	}
}

// close DB
void SQLite3ConnectionIntf::close()
{
	if (db != NULL) {
		db->close();
		db = NULL;
	}
}

// execute query (direct)
void SQLite3ConnectionIntf::executeQuery(tTJSString & query)
{
	if (db != NULL) {
		sqlite3plus::SqliteStatementPtr statement = db->createStatement(query.c_str());
		statement->execute();
	}
}

// check query
bool SQLite3ConnectionIntf::isCompleteSQL(tTJSString & query)
{
	bool rv = false;
	if (db != NULL) {
		rv = db->isCompleteSql(query.c_str());
	}
	return rv;
}

// column names
tjs_int SQLite3ConnectionIntf::columnNames(tTJSString & table, iTJSDispatch2 * array)
{
	tjs_int cols = 0;
	if (db != NULL) {
		std::vector<sqlite3plus::SqliteString> vnames = db->columnNames(table.c_str());
		std::vector<sqlite3plus::SqliteString>::iterator it;
		cols = vnames.size();
		for (it = vnames.begin(); it != vnames.end(); it++) {
			tTJSVariant name(it->c_str());
			Array_FuncCall_add(array, &name);
		}
	}
	return cols;
}

/*
SQLite3StatementIntf * SQLite3ConnectionIntf::createStatement(tTJSString & query)
{
	SQLite3StatementIntf * p = NULL;
	if (db != NULL) {
		p = new SQLite3StatementIntf();
		p->set(db, query);
	}
	return p;
}
*/

// get version (static)
void SQLite3ConnectionIntf::version(tTJSString & rv)
{
	rv = (sqlite3plus::SqliteConnection::version()).c_str();
}

// get connection ref.
sqlite3plus::SqliteConnection * SQLite3ConnectionIntf::getConnection()
{
	return db;
}



// class SQLite3StatementIntf

void SQLite3StatementIntf::set(SQLite3ConnectionIntf * c, tTJSString & query)
{
	conn = c;
	sqlite3plus::SqliteConnection * db = c->getConnection();
	sqlite3plus::SqliteStatementPtr st = db->createStatement(query.c_str());
	this->statement = st;
}

void SQLite3StatementIntf::execute()
{
	statement->execute();
}

void SQLite3StatementIntf::reset()
{
	statement->reset();
}

sqlite3plus::SqliteStatementPtr SQLite3StatementIntf::getStatement()
{
	return statement;
}





// class SQLite3CursorIntf

void SQLite3CursorIntf::set(SQLite3StatementIntf * s)
{
	statement = s;
	sqlite3plus::SqliteStatementPtr st = s->getStatement();
	sqlite3plus::SqliteCursorPtr cr = st->executeQuery();
	this->cursor = cr;
}

bool SQLite3CursorIntf::next()
{
	return cursor->next();
}

bool SQLite3CursorIntf::isEOR()
{
	return cursor->eor();
}

void SQLite3CursorIntf::reset()
{
	return cursor->reset();
}

tjs_int SQLite3CursorIntf::columnType(tjs_int index)
{
	sqlite3plus::SqliteCursorPtr c = cursor;
	if (c->isInt(index))  return TP_SQ3_DATATYPE_INTEGER;
	if (c->isStr(index))  return TP_SQ3_DATATYPE_STRING;
	if (c->isReal(index)) return TP_SQ3_DATATYPE_REAL;
	if (c->isNull(index)) return TP_SQ3_DATATYPE_NULL;
	if (c->isBlob(index)) return TP_SQ3_DATATYPE_BLOB;
	return TP_SQ3_DATATYPE_VOID;
}


sqlite3plus::SqliteCursorPtr SQLite3CursorIntf::getCursor()
{
	return cursor;
}

tjs_int SQLite3CursorIntf::columnNum()
{
	return cursor->colNum();
}

void SQLite3CursorIntf::columnName(tjs_int index, tTJSString & rv)
{
	rv = cursor->colName(index).c_str();
}


