//---------------------------------------------------------------------------
#include "util_sqlite.h"
#include "KRObjIntfEx.h"
#include "SQLite3Intf.h"
#include "NIHolder.h"
//---------------------------------------------------------------------------

#define SQ3_EXCEPTION_HEADER TJS_W("TVP Plugin: SQLite3:")
#define SQ3_THROW(_e)         TVPThrowExceptionMessage(SQ3_EXCEPTION_HEADER TJS_W(" %1"), _e.reason());
#define SQ3_THROW_INVALID     TVPThrowExceptionMessage(SQ3_EXCEPTION_HEADER TJS_W(" Invalid Object"));
#define SQ3_THROW_CREATENEW   TVPThrowExceptionMessage(SQ3_EXCEPTION_HEADER TJS_W(" You MUST NOT create new instance explicitly"));
#define SQ3_THROW_UNSUPPORTED TVPThrowExceptionMessage(SQ3_EXCEPTION_HEADER TJS_W(" Unsupported data type"));

#define SQ3_CATCH(_e) catch (sqlite3plus::SqliteException _e)
#define SQ3_CONSTRUCTOR_MAGIC_NUMBER      602136489
#define SQ3_CONSTRUCTOR_MAGIC_NUMBER_STR "602136489"

// SQLite3ijlCeBuCX^XێB
// ANZX̓CfbNXi0ȏ̐jōsB
static TPNativeInstanceHolder<SQLite3ConnectionIntf *> connNI;
static TPNativeInstanceHolder<SQLite3StatementIntf *>  statementNI;
static TPNativeInstanceHolder<SQLite3CursorIntf *>     cursorNI;

/*
lCeBuCX^X͎gIDƎqCX^XIDxN^B
g̃N[YюSɁAqCX^XċAIɎEdg݂B
*/

// [Ȏȁ[A

// Xe[gg̃N[YAfXgNVŎgB
static void killChildCursors(SQLite3StatementIntf * st)
{
	if (st == NULL) return;
	std::vector<tjs_int>::iterator it;
	for (it = st->childID.begin(); it != st->childID.end(); it++) {
		cursorNI.erase(*it);
	}
}
// RlNṼN[YAfXgNVŎgB
static void killChildStatements(SQLite3ConnectionIntf * db)
{
	if (db == NULL) return;
	std::vector<tjs_int>::iterator it;
	for (it = db->childID.begin(); it != db->childID.end(); it++) {
		SQLite3StatementIntf * st = statementNI.at(*it);
		killChildCursors(st);
		statementNI.erase(*it);
	}
}


// 
#define BEGIN_STD_ENI_PROCESS(_obj) \
	if (_obj != NULL) { \
		try {

#define END_STD_ENI_PROCESS \
		} SQ3_CATCH(e) { \
			SQ3_THROW(e) \
		} \
	} else { \
		SQ3_THROW_INVALID \
	}






//---------------------------------------------------------------------------
// {Iɂ͉ȂĂǁc
//---------------------------------------------------------------------------
void TVPCallBeforeV2LinkExit(void)
{
}

//---------------------------------------------------------------------------



// Connection

//---------------------------------------------------------------------------
// pseudo-constructor (callback)
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3ClassWhenCreateNewFunction)
{
	TP_REQ_NUM_PARAMS(0)

	tjs_int id = connNI.nextID();
	connNI.setID(objthis, id);
	SQLite3ConnectionIntf * db = new SQLite3ConnectionIntf();
	connNI.push_back(db);
	db->instanceID = id;

	TP_SET_RETURN_VALUE_INT(id)
	return TJS_S_OK;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// pseudo-destructor (callback)
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3ClassWhenInvalidateFunction)
{
	TP_REQ_NUM_PARAMS(0)

	tjs_int id = connNI.getID(objthis);
	SQLite3ConnectionIntf * db = connNI.at(id);

	// qE
	killChildStatements(db);
	// ijn
	if (db != NULL) {
		db->close();
	}
	connNI.erase(id);

	return TJS_S_OK;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// open
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3OpenFunction)
{
	TP_REQ_NUM_PARAMS(1)
	TP_CHK_PARAM_MUST_STR
	TP_GET_REQ_PARAM_STR(storage)

	tjs_int rv = 0;

	SQLite3ConnectionIntf * db = connNI.get(objthis);
	BEGIN_STD_ENI_PROCESS(db)
		db->open(storage);
		rv = 1;
	END_STD_ENI_PROCESS

	TP_SET_RETURN_VALUE_INT(rv)
	return TJS_S_OK;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// close
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3CloseFunction)
{
	TP_REQ_NUM_PARAMS(0)

	tjs_int id = connNI.getID(objthis);
	SQLite3ConnectionIntf * db = connNI.at(id);

	// qE
	killChildStatements(db);
	// ijn
	if (db != NULL) {
		db->close();
	}
	connNI.erase(id);

	return TJS_S_OK;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// execute query
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3ExecuteQueryFunction)
{
	TP_REQ_NUM_PARAMS(1)
	TP_CHK_PARAM_MUST_STR
	TP_GET_REQ_PARAM_STR(query)

	tjs_int rv = 0;

	SQLite3ConnectionIntf * db = connNI.get(objthis);
	BEGIN_STD_ENI_PROCESS(db)
		db->executeQuery(query);
		rv = 1;
	END_STD_ENI_PROCESS

	TP_SET_RETURN_VALUE_INT(rv)
	return TJS_S_OK;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// create statement
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3CreateStatementFunction)
{
	TP_REQ_NUM_PARAMS(1)
	TP_CHK_PARAM_MUST_STR
	TP_GET_REQ_PARAM_STR(query)

	SQLite3ConnectionIntf * db = connNI.get(objthis);
	if (db != NULL && result) {
		// y`
		TVPExecuteExpression(TJS_W("new SQLite3Statement((int)" SQ3_CONSTRUCTOR_MAGIC_NUMBER_STR ");"), result);
		iTJSDispatch2 * newobj = result->AsObjectNoAddRef();
		SQLite3StatementIntf * st = statementNI.get(newobj);
		st->set(db, query);
		// qIDo^
		db->childID.push_back(st->instanceID);
	} else if (result) {
		result->Clear(); // void
	} else {
		SQ3_THROW_INVALID
	}

	return TJS_S_OK;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// check query
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3IsCompleteSQLFunction)
{
	TP_REQ_NUM_PARAMS(1)
	TP_CHK_PARAM_MUST_STR
	TP_GET_REQ_PARAM_STR(query)

	bool rv = false;

	SQLite3ConnectionIntf * db = connNI.get(objthis);
	if (db != NULL) {
		rv = db->isCompleteSQL(query);
	} else {
		SQ3_THROW_INVALID
	}

	TP_SET_RETURN_VALUE_INT(rv)
	return TJS_S_OK;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// column names
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3ColumnNamesFunction)
{
	TP_REQ_NUM_PARAMS(1)
	TP_CHK_PARAM_MUST_STR
	TP_GET_REQ_PARAM_STR(table)

	bool rv = false;

	SQLite3ConnectionIntf * db = connNI.get(objthis);
	if (db != NULL) {
		if (result) {
			tjs_int cols = 0;
			iTJSDispatch2 * names = TJSCreateArrayObject();
			db->addInternalObject(names);
			try {
				cols = db->columnNames(table, names);
			} SQ3_CATCH(e) {
				SQ3_THROW(e)
			}
			TP_SET_RETURN_VALUE_OBJ(names)
			names->Release();
		}
	} else {
		SQ3_THROW_INVALID
	}

	return TJS_S_OK;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// static version
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3ClassVersionFunction)
{
	TP_REQ_NUM_PARAMS(0)

	tTJSString ver("");
	SQLite3ConnectionIntf::version(ver);

	TP_SET_RETURN_VALUE_PTR(ver)

	return TJS_S_OK;
}
//---------------------------------------------------------------------------




// Statement

//---------------------------------------------------------------------------
// pseudo-constructor (callback)
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3StatementClassWhenCreateNewFunction)
{
	TP_REQ_NUM_PARAMS(0)

/*
TODO:
̃\bh́ASQLite3StatementNX̃RXgN^Ă΂B
̃RXgN^́A[Uɂ閾Inew邱Ƃ҂ĂȂB
I new SQLite3Statement() ꂽƂɂ́AO𓊂@\Kvł낤B
*/
	TP_GET_OPT_PARAM_INT(magic, 0)
	if (magic != SQ3_CONSTRUCTOR_MAGIC_NUMBER) {
		SQ3_THROW_CREATENEW
	}

	tjs_int id = statementNI.nextID();
	statementNI.setID(objthis, id);
	SQLite3StatementIntf * st = new SQLite3StatementIntf();
	statementNI.push_back(st);
	st->instanceID = id;

	TP_SET_RETURN_VALUE_INT(id)
	return TJS_S_OK;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// pseudo-destructor (callback)
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3StatementClassWhenInvalidateFunction)
{
	TP_REQ_NUM_PARAMS(0)

	tjs_int id = statementNI.getID(objthis);
	SQLite3StatementIntf * st = statementNI.at(id);
	killChildCursors(st);
	statementNI.erase(id);

	return TJS_S_OK;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// close
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3StatementCloseFunction)
{
	TP_REQ_NUM_PARAMS(0)

	tjs_int id = statementNI.getID(objthis);
	SQLite3StatementIntf * st = statementNI.at(id);
	killChildCursors(st);
	statementNI.erase(id);

	return TJS_S_OK;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// reset
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3StatementResetFunction)
{
	TP_REQ_NUM_PARAMS(0)

	SQLite3StatementIntf * st = statementNI.get(objthis);
	BEGIN_STD_ENI_PROCESS(st)
		st->reset();
	END_STD_ENI_PROCESS

	return TJS_S_OK;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// execute
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3StatementExecuteFunction)
{
	TP_REQ_NUM_PARAMS(0)

	tjs_int rv = 0;

	SQLite3StatementIntf * st = statementNI.get(objthis);
	BEGIN_STD_ENI_PROCESS(st)
		st->execute();
		rv = 1;
	END_STD_ENI_PROCESS

	TP_SET_RETURN_VALUE_INT(rv)
	return TJS_S_OK;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// set param
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3StatementSetParamFunction)
{
	TP_REQ_NUM_PARAMS(2)
	TP_GET_REQ_PARAM_INT(index)

	tjs_int rv = 0;

	SQLite3StatementIntf * st = statementNI.get(objthis);
	if (st != NULL) {
		// TP_GET_REQ_PARAM_UINT(_lv) tjs_uint32 _lv = (tTVInteger)*param[tp_param_idx++];
		tTJSVariant * par = param[__tp_param_idx]; // __tp_param_idxQƂ̂͂낵Ȃc
		tjs_int parType = par->Type();
		sqlite3plus::SqliteStatementPtr nst = st->getStatement();
		try {
			switch (parType) {
			case tvtInteger:
				{
					TP_GET_REQ_PARAM_INT(var)
					nst->setParam(index, var);
				}
				break;
			case tvtReal:
				{
					TP_GET_REQ_PARAM_REAL(var)
					nst->setParam(index, var);
				}
				break;
			case tvtString:
				{
					TP_GET_REQ_PARAM_STR(var)
					nst->setParam(index, var.c_str());
				}
				break;
			case tvtVoid:
				break;
			case tvtObject:
				break;
			case tvtOctet:
				SQ3_THROW_UNSUPPORTED
				break;
			default:
				break;
			}
			rv = 1;
		} SQ3_CATCH(e) {
			SQ3_THROW(e)
		}
	} else {
		SQ3_THROW_INVALID
	}

	TP_SET_RETURN_VALUE_INT(rv)

	return TJS_S_OK;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// open cursor
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3StatementOpenCursorFunction)
{
	TP_REQ_NUM_PARAMS(0)

	SQLite3StatementIntf * st = statementNI.get(objthis);
	if (st != NULL && result) {
		// y`
		TVPExecuteExpression(TJS_W("new SQLite3Cursor((int)" SQ3_CONSTRUCTOR_MAGIC_NUMBER_STR ");"), result);
		iTJSDispatch2 * newobj = result->AsObjectNoAddRef();
		SQLite3CursorIntf * cr = cursorNI.get(newobj);
		cr->set(st);
		// qIDo^
		st->childID.push_back(cr->instanceID);
	} else if (result) {
		result->Clear(); // void
	} else {
		SQ3_THROW_INVALID
	}

	return TJS_S_OK;
}
//---------------------------------------------------------------------------




// Cursor

//---------------------------------------------------------------------------
// pseudo-constructor (callback)
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3CursorClassWhenCreateNewFunction)
{
	TP_REQ_NUM_PARAMS(0)

/*
TODO:
̃\bh́ASQLite3CursorNX̃RXgN^Ă΂B
̃RXgN^́A[Uɂ閾Inew邱Ƃ҂ĂȂB
I new SQLite3Cursor() ꂽƂɂ́AO𓊂@\Kvł낤B
*/
	TP_GET_OPT_PARAM_INT(magic, 0)
	if (magic != SQ3_CONSTRUCTOR_MAGIC_NUMBER) {
		SQ3_THROW_CREATENEW
	}

	tjs_int id = cursorNI.nextID();
	cursorNI.setID(objthis, id);
	SQLite3CursorIntf * cr = new SQLite3CursorIntf();
	cursorNI.push_back(cr);
	cr->instanceID = id;

	TP_SET_RETURN_VALUE_INT(id)
	return TJS_S_OK;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// pseudo-destructor (callback)
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3CursorClassWhenInvalidateFunction)
{
	TP_REQ_NUM_PARAMS(0)

	cursorNI.erase(objthis);

	return TJS_S_OK;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// next
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3CursorNextFunction)
{
	TP_REQ_NUM_PARAMS(0)
	bool rv = false;

	SQLite3CursorIntf * cr = cursorNI.get(objthis);
	if (cr != NULL) {
		rv = cr->next();
	} else {
		SQ3_THROW_INVALID
	}

	TP_SET_RETURN_VALUE_INT(rv)
	return TJS_S_OK;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// isEOR
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3CursorIsEORFunction)
{
	TP_REQ_NUM_PARAMS(0)
	bool rv = false;

	SQLite3CursorIntf * cr = cursorNI.get(objthis);
	if (cr != NULL) {
		rv = cr->isEOR();
	} else {
		SQ3_THROW_INVALID
	}

	TP_SET_RETURN_VALUE_INT(rv)
	return TJS_S_OK;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// reset
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3CursorResetFunction)
{
	TP_REQ_NUM_PARAMS(0)

	SQLite3CursorIntf * cr = cursorNI.get(objthis);
	if (cr != NULL) {
		cr->reset();
	} else {
		SQ3_THROW_INVALID
	}

	return TJS_S_OK;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// value
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3CursorValueFunction)
{
	TP_REQ_NUM_PARAMS(1)
	TP_GET_REQ_PARAM_INT(index)
	TP_GET_OPT_PARAM_INT(type, TP_SQ3_DATATYPE_AUTO)

	SQLite3CursorIntf * cr = cursorNI.get(objthis);
	if (cr != NULL) {
		try {
			type = (type == TP_SQ3_DATATYPE_AUTO) ? cr->columnType(index) : type;
		} SQ3_CATCH(e) {
			SQ3_THROW(e)
		}
		sqlite3plus::SqliteCursorPtr ncr = cr->getCursor();
		switch (type) {
		case TP_SQ3_DATATYPE_INTEGER:
			TP_SET_RETURN_VALUE_INT(ncr->valueAsInt64(index))
			break;
		case TP_SQ3_DATATYPE_STRING:
			TP_SET_RETURN_VALUE_PTR(ncr->valueAsStr(index).c_str())
			break;
		case TP_SQ3_DATATYPE_REAL:
			TP_SET_RETURN_VALUE_REAL(ncr->valueAsReal(index))
			break;
		case TP_SQ3_DATATYPE_NULL:
			{
				tTJSVariant null((iTJSDispatch2 *)NULL);
				TP_SET_RETURN_VALUE_VAR(null)
			}
			break;
		case TP_SQ3_DATATYPE_BLOB:
		default:
			SQ3_THROW_UNSUPPORTED
		}
	} else {
		SQ3_THROW_INVALID
	}

	return TJS_S_OK;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// column num
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3CursorColumnNumFunction)
{
	TP_REQ_NUM_PARAMS(0)
	tjs_int rv = 0;

	SQLite3CursorIntf * cr = cursorNI.get(objthis);
	if (cr != NULL) {
		rv = cr->columnNum();
	} else {
		SQ3_THROW_INVALID
	}
	TP_SET_RETURN_VALUE_INT(rv)
	return TJS_S_OK;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// column types
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3CursorColumnTypesFunction)
{
	TP_REQ_NUM_PARAMS(0)
	tjs_int rv = 0;

	SQLite3CursorIntf * cr = cursorNI.get(objthis);
	if (cr != NULL) {
		if (result) {
			iTJSDispatch2 * types = TJSCreateArrayObject();
			cr->addInternalObject(types);
			tjs_int num = cr->columnNum();
			for (tjs_int i = 0; i < num; i++) {
				tjs_int type = TP_SQ3_DATATYPE_VOID;
				try {
					type = cr->columnType(i);
				} SQ3_CATCH(e) {
					SQ3_THROW(e)
				}
				tTJSVariant val(type);
				Array_FuncCall_add(types, &val);
			}
			TP_SET_RETURN_VALUE_OBJ(types)
			types->Release();
		}
	} else {
		SQ3_THROW_INVALID
	}

	return TJS_S_OK;
}
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
// column names
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3CursorColumnNamesFunction)
{
	TP_REQ_NUM_PARAMS(0)
	tjs_int rv = 0;

	SQLite3CursorIntf * cr = cursorNI.get(objthis);
	if (cr != NULL) {
		if (result) {
			iTJSDispatch2 * names = TJSCreateArrayObject();
			cr->addInternalObject(names);
			tjs_int num = cr->columnNum();
			for (tjs_int i = 0; i < num; i++) {
				tTJSString name;
				try {
					cr->columnName(i, name);
				} SQ3_CATCH(e) {
					SQ3_THROW(e)
				}
				tTJSVariant val(name);
				Array_FuncCall_add(names, &val);
			}
			TP_SET_RETURN_VALUE_OBJ(names)
			names->Release();
		}
	} else {
		SQ3_THROW_INVALID
	}

	return TJS_S_OK;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// close
//---------------------------------------------------------------------------
TVP_PLUGIN_FUNCCALL_DECL(tSQLite3CursorCloseFunction)
{
	TP_REQ_NUM_PARAMS(0)

	cursorNI.erase(objthis);

	return TJS_S_OK;
}
//---------------------------------------------------------------------------

