/**************************************************************
*
*  Jul. 8, 2004
*
*  The author disclaims copyright to this source code.
*  In place of a legal notice, here is a blessing:
*
*    May you do good and not evil.
*    May you find forgiveness for yourself and forgive others.
*    May you share freely, never taking more than you give.
*
**************************************************************/
#ifndef SqliteConnection_H
#define SqliteConnection_H


#include <sqlite3.h>
#include <vector>
#include "SqliteException.h"
#include "SqliteString.h"
#include "SqliteSmartPtr.h"
#include "SqliteStatement.h"

namespace sqlite3plus {

/// SQLitef[^x[XƂ̐ڑbvNX
class SqliteConnection
{
protected:
	///x[XƂȂSQLiteDBnh
	sqlite3*			m_db;

public:
	///f[^x[Xt@CJɍ\z
	SqliteConnection(void)
		: m_db(NULL)
	{}

	/// f[^x[Xt@Cw肵č\z
	/**
	 * @param dbName f[^x[XƂĊJt@C̖O
	 * @throw ConnectionException I[vłȂ
	 */
	SqliteConnection(const _TCHAR* dbName)
		: m_db(NULL)
	{
		open(dbName);
	}

	virtual ~SqliteConnection(void)
	{
		close();
	}

	/// DBnh擾
	/**
	 * @return SQLiteDBnh
	 */
	sqlite3* db() const
	{
		return m_db;
	}

	///w肵f[^x[Xt@CJ
	/**
	 * @param dbName f[^x[XƂĊJt@C̖O
	 * @throw ConnectionException I[vłȂ
	 */
	void open(const _TCHAR* dbName)
	{
		SqliteStringW fileName = SqliteString(dbName).toUTF16();

		int errCode = sqlite3_open16(fileName.c_str(), &m_db);

		if(SQLITE_OK != errCode)
		{
			if(m_db)
			{
				SqliteStringW errMsg = (const wchar_t*)sqlite3_errmsg16(m_db);
				throw ConnectionException(errMsg.toANSI().c_str());
			}
			else
			{
				SqliteString errMsg = _T("Can't open ");
				errMsg += dbName;
				throw ConnectionException(errMsg.toANSI().c_str());
			}
		}
	}

	/// JĂf[^x[X
	void close()
	{
		if(m_db)
		{
			sqlite3_close(m_db);
			m_db = NULL;
		}
	}

	/// }`Xbh쎞Aʂ̃XbhsĂXe[ggIB
	/**
	 * @note Xe[ggsɊ荞܂ꂽXbhɂ́AG[R[hSQLITE_INTERRUPT
	 * QueryException
	 */
	void interrupt()
	{
		if(m_db)
		{
			::sqlite3_interrupt(m_db);
		}
	}

	/// Xe[gg\z
	/**
	 * @param sql sׂSQL
	 * @return \zꂽXe[ggIuWFNgւ̃X}[g|C^
	 */
	SqliteStatementPtr createStatement(const _TCHAR* sql)
	{
		// prepare statement
		sqlite3_stmt* stmt = NULL;
		const void* unused = NULL;
		int res = SQLITE_OK;

		SqliteStringW strSql = SqliteString(sql).toUTF16();

		res = sqlite3_prepare16(m_db,
			strSql.toUTF16().c_str(),
			-1, // strSql.toUTF16().c_str() is null terminated
			&stmt,
			&unused);

		if(SQLITE_OK == res)
			return SqliteStatementPtr(new SqliteStatement(sql, stmt, m_db));
		else
		{
			SqliteStringW errMsg = (const wchar_t*)sqlite3_errmsg16(m_db);
			throw QueryException(errMsg.toANSI().c_str(), res);
		}
	}

	/// SQLitẽo[W擾
	/**
	 * @return SQLitẽo[W
	 */
	static SqliteString version()
	{
		return SqliteString::fromANSI(sqlite3_libversion());
	}

	/// Ōɑ}j[NȍsID擾
	/**
	 * @return Ōɑ}j[NȍsID
	 */
	__int64 lastInsertRowID()
	{
		return sqlite3_last_insert_rowid(m_db);
	}

	/// ŌINSERT, DELETE, UPDATE̎sŕύXꂽs̐擾
	/**
	 * @return ύXꂽs
	 * @note WHERE߂̂Ȃ DELETÉASQLiteDROP TABLECREATE TABLE
	 * s邱ƂɂĎĂB̏ꍇAlastChangedRowNum()
	 * K0ԂB
	 */
	int lastChangedRowNum()
	{
		return ::sqlite3_changes(m_db);
	}

	/// f[^x[XJĂύXꂽs擾
	/**
	 * @return f[^x[XJĂύXꂽs
	 */
	int lastChangedRowNumTotal()
	{
		return sqlite3_total_changes(m_db);
	}

	///w肵񂪁ASQLƂĐĂ邩ǂ`FbN
	/**
	 * @return SQLƂĐĂtrueAsSȂfalse
	 */
	bool isCompleteSql(const _TCHAR* sql)
	{
		SqliteStringW strSql = SqliteString(sql).toUTF16();
		return (sqlite3_complete16(strSql.c_str()) != 0);
	}

	/// we[u̗񖼔z擾
	/**
	 * @param tableName 񖼂擾e[u̖O
	 * @return 񖼂̔z
	 */
	std::vector<SqliteString> columnNames(const _TCHAR* tableName)
	{
		std::vector<SqliteString> ret;

		SqliteString sql = _T("SELECT * FROM '");
		sql.append(tableName);
		sql.append(_T("' WHERE 0;"));

		SqliteStatementPtr st = createStatement(sql.c_str());
		SqliteCursorPtr cur = st->executeQuery();
		cur->next();

		for(int i=0; i<cur->colNum(); i++)
		{
			ret.push_back(cur->colName(i));
		}

		return ret;
	}
};


}; //end of namespace sqlite3plus

#endif SqliteConnection_H

