/**************************************************************
*
*  Jul. 8, 2004
*
*  The author disclaims copyright to this source code.
*  In place of a legal notice, here is a blessing:
*
*    May you do good and not evil.
*    May you find forgiveness for yourself and forgive others.
*    May you share freely, never taking more than you give.
*
**************************************************************/
#ifndef SqliteStatement_H
#define SqliteStatement_H


#include <sqlite3.h>
#include "SqliteString.h"
#include "SqliteSmartPtr.h"

namespace sqlite3plus {

/// J[\
class SqliteCursor
{
	friend class SqliteStatement;

protected:
	/// UgZbg̑SĂ̍s擾IƁAtrueɃZbg
	bool					m_eor;
	/// ݂܂łɎ擾s̐
	int						m_rowNum;
	/// 
	int						m_colNum;
	/// f[^x[Xnh
	sqlite3*				m_db;
	/// Xe[ggnh
	sqlite3_stmt*			m_stmt;

public:
	/// f[^x[Xnh擾
	sqlite3* db()
	{
		return m_db;
	}

	/// Xe[ggnh擾
	sqlite3_stmt* stmt()
	{
		return m_stmt;
	}

protected:
	/// RXgN^B[U͒ڃJ[\NX\zĂ͂ȂȂB
	SqliteCursor(sqlite3_stmt* stmt, sqlite3* db)
		: m_db(db)
		, m_stmt(stmt)
		, m_rowNum(0)
		, m_eor(false)
	{
		m_colNum = sqlite3_column_count(m_stmt);
	}

public:
	///fXgN^
	virtual ~SqliteCursor()
	{}

	/// UgZbg̑SĂ̍s擾IĂ邩ǂ𔻒
	/**
	 * @return ȏ㌋ʍsȂtrueA܂݂false
	 */
	bool eor()
	{
		return m_eor;
	}

	/// ݂܂łɎ擾sԂ
	/**
	 * @return ݂܂łɎ擾s
	 */
	int rowNum()
	{
		return m_rowNum;
	}

	/// UgZbg̗񐔂Ԃ
	/**
	 * @return 
	 */
	int colNum()
	{
		return m_colNum;
	}

	/// ̍s擾
	/**
	 * @return Ɏ̌ʍstrueAꂪŌ̍sȂfalse
	 * @throw QueryException ʂ̎擾Ɏsꍇ
	 */
	bool next()
	{
		int res = sqlite3_step(m_stmt);

		if(SQLITE_ROW == res)
		{
			m_rowNum++;
			return true;
		}
		else if(SQLITE_DONE == res)
		{
			m_eor = true;
			return false;
		}
		else
		{
			checkResult(res);
			return false;
		}
	}

	/// J[\ʒuďAUgZbg̐擪ɖ߂
	/**
	 * @throw QueryException J[\̍ďɎsꍇ
	 * @note SqliteStatement::reset()Ɠ@\BǂsĂʂłB
	 */
	void reset()
	{
		int res = sqlite3_reset(m_stmt);
		checkResult(res);
	}

	/// ݃|CgĂs̎w̒lNULLǂ肷
	/**
	 * @param index `FbN̔ԍi0n܂j
	 * @return NULLȂtrueAłȂfalse
	 * @throw OutOfBoundsException index͈͂EĂꍇ
	 */
	bool isNull(int index)
	{
		checkColIndex(index, "isNull");
		return (sqlite3_column_type(m_stmt, index) == SQLITE_NULL);
	}

	/// ݃|CgĂs̎w̃^Cv񂩂ǂ肷
	/**
	 * @param index `FbN̔ԍi0n܂j
	 * @return ȂtrueAłȂfalse
	 * @throw OutOfBoundsException index͈͂EĂꍇ
	 */
	bool isStr(int index)
	{
		checkColIndex(index, "isStr");
		return (sqlite3_column_type(m_stmt, index) == SQLITE_TEXT);
	}

	/// ݃|CgĂs̎w̃^Cvǂ肷
	/**
	 * @param index `FbN̔ԍi0n܂j
	 * @return ȂtrueAłȂfalse
	 * @throw OutOfBoundsException index͈͂EĂꍇ
	 */
	bool isInt(int index)
	{
		checkColIndex(index, "isInt");
		return (sqlite3_column_type(m_stmt, index) == SQLITE_INTEGER);
	}

	/// ݃|CgĂs̎w̃^Cvǂ肷
	/**
	 * @param index `FbN̔ԍi0n܂j
	 * @return ȂtrueAłȂfalse
	 * @throw OutOfBoundsException index͈͂EĂꍇ
	 */
	bool isReal(int index)
	{
		checkColIndex(index, "isReal");
		return (sqlite3_column_type(m_stmt, index) == SQLITE_FLOAT);
	}

	/// ݃|CgĂs̎w̃^CvBLOBǂ肷
	/**
	 * @param index `FbN̔ԍi0n܂j
	 * @return BLOBȂtrueAłȂfalse
	 * @throw OutOfBoundsException index͈͂EĂꍇ
	 */
	bool isBlob(int index)
	{
		checkColIndex(index, "isBlob");
		return (sqlite3_column_type(m_stmt, index) == SQLITE_BLOB);
	}

	/// ݃|CgĂs̎w̒l𕶎ƂĎ擾
	/**
	 * @param index ԍi0n܂j
	 * @return 擾l
	 * @throw OutOfBoundsException index͈͂EĂꍇ
	 */
	SqliteString valueAsStr(int index)
	{
		checkColIndex(index, "valueAsStr");
		return SqliteString::fromUTF16(
			(const wchar_t*)sqlite3_column_text16(m_stmt, index));
	}

	/// ݃|CgĂs̎w̒l𐮐lƂĎ擾
	/**
	 * @param index ԍi0n܂j
	 * @return 擾l
	 * @throw OutOfBoundsException index͈͂EĂꍇ
	 */
	int valueAsInt(int index)
	{
		checkColIndex(index, "valueAsInt");
		return sqlite3_column_int(m_stmt, index);
	}

	/// ݃|CgĂs̎w̒l64bitlƂĎ擾
	/**
	 * @param index ԍi0n܂j
	 * @return 擾l
	 * @throw OutOfBoundsException index͈͂EĂꍇ
	 */
	__int64 valueAsInt64(int index)
	{
		checkColIndex(index, "valueAsInt64");
		return sqlite3_column_int64(m_stmt, index);
	}

	/// ݃|CgĂs̎w̒llƂĎ擾
	/**
	 * @param index ԍi0n܂j
	 * @return 擾l
	 * @throw OutOfBoundsException index͈͂EĂꍇ
	 */
	double valueAsReal(int index)
	{
		checkColIndex(index, "valueAsReal");
		return sqlite3_column_double(m_stmt, index);
	}

	/// BLOBl̐擪AhXԂ
	/**
	 * @param index ԍi0n܂j
	 * @return 擾AhX
	 * @throw OutOfBoundsException index͈͂EĂꍇ
	 */
	const void* valueAsBlob(int index)
	{
		checkColIndex(index, "valueAsBlob");
		return sqlite3_column_blob(m_stmt, index);
	}

	/// l߂oCgTCYԂB
	/**
	 * @param index ԍi0n܂j
	 * @return oCg
	 * @throw OutOfBoundsException index͈͂EĂꍇ
	 * @note lłꍇAUTF8ɕϊɃoCgTCYZoB
	 */
	int valueBytes(int index)
	{
		checkColIndex(index, "valueBytes");
		return sqlite3_column_bytes(m_stmt, index);
	}

	/// l߂oCgTCYԂB
	/**
	 * @param index ԍi0n܂j
	 * @return oCg
	 * @throw OutOfBoundsException index͈͂EĂꍇ
	 * @note lłꍇAUTF16ɕϊɃoCgTCYZoB
	 */
	int valueBytes16(int index)
	{
		checkColIndex(index, "valueBytes16");
		return sqlite3_column_bytes16(m_stmt, index);
	}

	/// w̖O擾
	/**
	 * @param index ԍ (0n܂j
	 * @return 
	 * @throw OutOfBoundsException index͈͂EĂꍇ
	 */
	SqliteString colName(int index)
	{
		checkColIndex(index, "colName");
		return SqliteString::fromUTF16(
			(const wchar_t*)sqlite3_column_name16(m_stmt, index));
	}

protected:
	void checkResult(int rescode)
	{
		if(SQLITE_OK != rescode)
		{
			SqliteStringW errMsg = (const wchar_t*)sqlite3_errmsg16(m_db);
			throw QueryException(errMsg.toANSI().c_str(), rescode);
		}
	}

	void checkColIndex(int index, const char* reason)
	{
		if(index < 0 || index >= m_colNum)
		{
			throw OutOfBoundsException(reason);
		}
	}
};

/// SqliteCursorIuWFNgւ̃X}[g|C^
typedef SqliteSmartPtr<SqliteCursor> SqliteCursorPtr;


/// Xe[gg
class SqliteStatement
{
	friend class SqliteConnection;

protected:
	/// SQL
	SqliteString	m_sql;
	/// Xe[ggnh
	sqlite3_stmt*	m_stmt;
	/// SQLitef[^x[Xnh
	sqlite3*		m_db;

public:
	/// ێSQL擾
	SqliteString sql()
	{
		return m_sql;
	}

	/// Xe[ggnh擾
	sqlite3_stmt* stmt()
	{
		return m_stmt;
	}

	/// f[^x[Xnh擾
	sqlite3* db()
	{
		return m_db;
	}

protected:
	/// RXgN^BSqliteConnection݂̂p
	/**
	 * @param sql SQLBPrepared Statement邽߁AChJ[h"?"𗘗p\B
	 * @param stmt Xe[ggnh
	 * @param db f[^x[Xnh
	 */
	SqliteStatement(const _TCHAR* sql, sqlite3_stmt* stmt, sqlite3* db)
		: m_sql(sql)
		, m_stmt(stmt)
		, m_db(db)
	{}

public:
	virtual ~SqliteStatement()
	{
		if(m_stmt)
			sqlite3_finalize(m_stmt);
	}

	/// xexecute, executeQueryXe[ggAĂexecuteł悤ɏԂZbg
	/**
	 * @throw QueryException ZbgɎsꍇ
	 * @note Prepared Statement̏ꍇAreset()ĂChJ[hɃoChꂽl͈ȑÔ܂܂łB
	 * KvȂAreset()ɍăoCh邱ƂłB
	 */
	void reset()
	{
		int res = sqlite3_reset(m_stmt);
		checkResult(res);
	}

	/// ʂԂȂXe[ggsiINSERT, DELETE, UPDATE, CREATE TABLEȂǁj
	/**
	 * @throw QueryException sɎsꍇ
	 */
	void execute()
	{
		int res = sqlite3_step(m_stmt);
		if( SQLITE_ROW != res &&
			SQLITE_DONE != res )
			checkResult(res);
	}

	/// Xe[ggsAʂ擾邽߂̃J[\Ԃ
	/**
	 * @return J[\IuWFNgւ̃X}[g|C^
	 * @note ̎_ł͗O𓊂ȂBSqliteCursor::next()œB
	 */
	SqliteCursorPtr executeQuery()
	{
		return SqliteCursorPtr(new SqliteCursor(m_stmt, m_db));
	}

	/// Prepared Statement̃ChJ[hɁAۂ̒loCh
	/**
	 * @param index CfbNXi1n܂j
	 * @param val oChl
	 * @throw QueryException LȃCfbNX͈͂𒴉߂Ƃ
	 * @note ŏ̃ChJ[h̃CfbNX́A1ł邱ƂɒӁB
	 * 0ł͂ȂB܂ALȃCfbNX͈͂𒴉߂ƂɓOA
	 * OutOfBoundsExceptionł͂ȂAQueryExceptionł邱ƂɒӁB
	 *iSQLite{̂̃`FbN@\ɂ̂ł邽߁j
	 */
	void setParam(int index, int val)
	{
		int res = sqlite3_bind_int(m_stmt, index, val);
		checkResult(res);
	}

	/// Prepared Statement̃ChJ[hɁAۂ̒loCh
	/**
	 * @param index CfbNXi1n܂j
	 * @param val oChl
	 * @throw QueryException LȃCfbNX͈͂𒴉߂Ƃ
	 * @note ŏ̃ChJ[h̃CfbNX́A1ł邱ƂɒӁB
	 * 0ł͂ȂB܂ALȃCfbNX͈͂𒴉߂ƂɓOA
	 * OutOfBoundsExceptionł͂ȂAQueryExceptionł邱ƂɒӁB
	 *iSQLite{̂̃`FbN@\ɂ̂ł邽߁j
	 */
	void setParam(int index, __int64 val)
	{
		int res = sqlite3_bind_int64(m_stmt, index, val);
		checkResult(res);
	}

	/// Prepared Statement̃ChJ[hɁAۂ̒loCh
	/**
	 * @param index CfbNXi1n܂j
	 * @param val oChl
	 * @throw QueryException LȃCfbNX͈͂𒴉߂Ƃ
	 * @note ŏ̃ChJ[h̃CfbNX́A1ł邱ƂɒӁB
	 * 0ł͂ȂB܂ALȃCfbNX͈͂𒴉߂ƂɓOA
	 * OutOfBoundsExceptionł͂ȂAQueryExceptionł邱ƂɒӁB
	 *iSQLite{̂̃`FbN@\ɂ̂ł邽߁j
	 */
	void setParam(int index, double val)
	{
		int res = sqlite3_bind_double(m_stmt, index, val);
		checkResult(res);
	}

	/// Prepared Statement̃ChJ[hɁAۂ̒loCh
	/**
	 * @param index CfbNXi1n܂j
	 * @param val oChl
	 * @throw QueryException LȃCfbNX͈͂𒴉߂Ƃ
	 * @note ŏ̃ChJ[h̃CfbNX́A1ł邱ƂɒӁB
	 * 0ł͂ȂB܂ALȃCfbNX͈͂𒴉߂ƂɓOA
	 * OutOfBoundsExceptionł͂ȂAQueryExceptionł邱ƂɒӁB
	 *iSQLite{̂̃`FbN@\ɂ̂ł邽߁j
	 */
	void setParam(int index, const _TCHAR* val)
	{
		SqliteStringW str = SqliteString(val).toUTF16();
		int res = sqlite3_bind_text16(
			m_stmt, index, str.c_str(),
			(int)str.length()*sizeof(wchar_t),
			SQLITE_TRANSIENT);

		checkResult(res);
	}

	/// Prepared Statement̃ChJ[hɁAۂ̒loCh
	/**
	 * @param index CfbNXi1n܂j
	 * @param val oChl
	 * @param nBytes BLOB̃oCg
	 * @param makeCopy BLOBRs[ĕێꍇɂtrue
	 * @throw QueryException LȃCfbNX͈͂𒴉߂Ƃ
	 * @note ŏ̃ChJ[h̃CfbNX́A1ł邱ƂɒӁB
	 * 0ł͂ȂB܂ALȃCfbNX͈͂𒴉߂ƂɓOA
	 * OutOfBoundsExceptionł͂ȂAQueryExceptionł邱ƂɒӁB
	 *iSQLite{̂̃`FbN@\ɂ̂ł邽߁j
	 */
	void setParam(int index, const void* val, int nBytes, bool makeCopy=true)
	{
		int res = sqlite3_bind_blob(m_stmt, index, val, nBytes,
			makeCopy ? SQLITE_TRANSIENT : SQLITE_STATIC);

		checkResult(res);
	}

protected:
	void checkResult(int rescode)
	{
		if(SQLITE_OK != rescode)
		{
			SqliteStringW errMsg = (const wchar_t*)sqlite3_errmsg16(m_db);
			throw QueryException(errMsg.toANSI().c_str(), rescode);
		}
	}
};

/// SqliteStatementIuWFNgւ̃X}[g|C^
typedef SqliteSmartPtr<SqliteStatement> SqliteStatementPtr;


}; //end of namespace sqlite3plus


#endif SqliteStatement_H
