
# to TDF
use strict;

use utf8;
my $ENCODE_NAME = ':encoding(shift_jis)';
binmode(STDIN,  $ENCODE_NAME);
binmode(STDOUT, $ENCODE_NAME);
binmode(STDERR, $ENCODE_NAME);

my $CurTag = '';	# 処理中のタグ名
my @Texts = ();
my $DocStruct = {};
my @FunctionList = ();
my @FunctionListWithSYN = ();
my $CurName = '';	# 処理中のメソッド名
my $Headings = {};	# メソッド名と見出し番号の関連付け
my $HeadingCounter = 2;
my $PlugInName = 'NoName';

my $TDFDir = '../doc/tdf';
my $FileContent = {};
my $IndexFileName = 'index.tdf';
my $TitleName = '名無し';

# オンラインマニュアル
my %OnlineRoot = (
	'!TVP' => 'http://devdoc.kikyou.info/tvp/docs/kr2doc/contents',
	'!TJS' => 'http://devdoc.kikyou.info/tvp/docs/tjs2doc/contents',
);

my %HList = (
'TITLE'       => ['TITLE', ''],
'DECLARATION' => ['TITLE', ''],
#'ABSTRACT'    => ['P', '/P'],
'ABSTRACT'    => ['SECTION 概要', '/SECTION'],
'SYNOPSYS'    => ['P', '/P'],
'EXCEPTION'   => ['SECTION 例外', '/SECTION'],
'DESCRIPTION' => ['SECTION 詳細', '/SECTION'],
'EXAMPLE'     => ['SECTION 例', '/SECTION'],
'ARGUMENT'    => ['SECTION 引数', '/SECTION'],
'RETURN'      => ['SECTION 戻り値', '/SECTION'],
'SEE'      => ['SECTION 関連項目', '/SECTION'],
);

while (<STDIN>) {
	next if (/^#/);
	s/\n//;
	# タグ
	if (/^[A-Z]/s) {
		out($CurTag, \@Texts);
		$CurTag = $_;
		@Texts = ();
	}
	if (/^\t(.*)/) {
		my $t = $1;
		if ($t =~ /^\s*$/) {
			$t .= '<<CR>>';
		}
		push (@Texts, $t);
	}
}

out($CurTag, \@Texts);

while (my ($k, $v) = (each %{$FileContent})) {
	next if ($k eq '_');
	my $file = "${TDFDir}/" . getFileNameBase($k) . '.tdf';
	open (F, ">$file") or die("$file: $!");
	binmode(F, $ENCODE_NAME);
	print F $v;
	close (F);
}

outIndex();


sub getFileNameBase {
	my ($f) = @_;
	return "f_${f}";
}


sub outIndex {
	open (F, ">$TDFDir/$IndexFileName") or die("$TDFDir/$IndexFileName: $!");
	binmode(F, $ENCODE_NAME);
print F << "END";
$FileContent->{'_'}

SECTION 関数一覧
P
このプラグインをリンクすると、下記の関数が追加されます。
/P
UL
END
	foreach (@FunctionList) {
		my $fn = getFileNameBase($_) . '.html';
		print F "LI* $fn $_\n";
	}
print F << "END";
/UL
/SECTION
END
	close (F);
}



sub out {
	my ($tag, $texts) = @_;
	my $rv = '';
	my $text = join("\n", @$texts);
	my $prp = 0;

	my $btag = $HList{$tag}->[0];
	my $etag = $HList{$tag}->[1];

	if ($tag eq 'PLUGIN') {
		$PlugInName = $text;
		$TDFDir .= "/$PlugInName";
		#print STDERR $PlugInName;

	} elsif ($tag eq 'TITLE') {
		$rv = "$btag $text\n\n";
		$TitleName = $text;
		print $rv;

	} elsif ($tag eq 'SEE') {
		$rv = "$btag\nUL\n";
		my @funcs = split(/\t*\n\t*/, $text);
		foreach (@funcs) {
			my $link = createLink("$_ -");
			$rv .= "LI* ${link}\n" if ($link);
		}
		$rv .= "$etag\n";

	} elsif ($tag eq 'DECLARATION') {
		my $static = $text =~ s/^\s*static\s+//;
		my $othercontext = $text =~ s/^\s*othercontext\s+//;

		$text =~ /(^.*)\((.*)\)/;
		my ($name, $args) = ($1, $2);
		$CurName = $name;
		my ($ctx, $member, $ctor) = ('global', $name, '');
		my $ctx_additional = '';
		my $global = 1;
		# with context (except global)
		if ($name =~ /\./) {
			($ctx, $member) = split(/\./, $name, 2);
			$ctor = ($member eq $ctx) ? 1 : '';
			# constructor
			if ($ctor) {
				$ctx = "global";
				$ctor = "new $ctx.";
			# others
			} else {
				$ctx = "global.$ctx";
				$ctx_additional = ($static) ? 
					' (static)' :
					' インスタンス';
			}
			$global = 0;
		}
		# with arguments
		if ($args) {
			my @aa = split(/\s*,\s*/, $args);
			my @ar = ();
			foreach my $t (@aa) {
				my $id = getArgID($CurName, $t);
				push @ar, "VERB <a href='#$id' title='$t : $name'>$t</a>";
			}
			$args = "\n" . join("\n, \n", @ar) . "\n";
		}
		# how to call
		my $callcode;
		if ($othercontext) {
			$callcode = "($ctx.$member incontextof\nVAR object\n)($args )";
		} elsif ($static) {
			$callcode = "$ctx.$member ($args )";
		} elsif ($ctor) {
			$callcode = "$ctor$member ($args )";
		} elsif ($global) {
			$callcode = "$ctx.$member($args )";
		} else {
			$callcode = "VAR instance\n.$member($args )";
		}

		push @FunctionList, $name;

		$rv = << "END";
$btag $name （<!--SYNOPSYS-->）

UL id="navi" title="ナビゲーション"
LI* index.html プラグイン $PlugInName 概要
LI* index.html#h2. 関数一覧
/UL

SECTION 文法
DL class='funcabstract'
DT メンバ所有者
DD+
ICODE $ctx$ctx_additional
/DD
DT 呼び出し
DD+
ICODE+
$callcode
/ICODE
/DD
/DL
/SECTION
END

	} elsif ($tag eq 'ARGUMENT') {
		$text = getArgList($texts);
		$text = "DL class='arglist'\n$text\n/DL\n";
		$text = asTDSIL($text);
		$rv = "$btag\n$text\n$etag\n\n";

	} elsif ($tag eq 'ABSTRACT') {
		$text = replaceCR($text, 'P');
		$text = asTDS($text, 'P');
		$text = asTDSIL($text);
		$rv = << "END";
$btag
P
$text
/P
$etag

END

	} elsif ($tag eq 'SYNOPSYS') {
		$rv = << "END";
$btag
$text
$etag

END

		$DocStruct->{'DECLARATION'} =~ s/<!--SYNOPSYS-->/$text/;

	} elsif ($tag eq 'DESCRIPTION' || $tag eq 'EXCEPTION') {
		$text = replaceCR($text, 'P');
		$text = asTDS($text, 'P');
		$text = asTDSIL($text);
		$rv = << "END";
$btag
P
$text
/P
$etag

END

	} elsif ($tag eq 'EXAMPLE') {
		$text = replaceCR($text, 'P');
		$text = asTDS($text, 'P');
		$text = asTDSIL($text);
		$rv = << "END";
$btag
P
$text
/P
$etag

END

	} elsif ($tag eq 'RETURN') {
		$prp = 1;
		$text = replaceCR($text, 'P');
		$text = asTDS($text, 'P');
		$text = asTDSIL($text);
		$rv = << "END";
$btag
P
$text
/P
$etag
$HList{'DECLARATION'}->[1]

END
	}

	$rv = removeEmptyPair($rv, 'P');
	$rv = removeEmptyPair($rv, 'DD');

	$DocStruct->{$tag} = $rv;

	# 関数のページ
	if ($prp) {
		$FileContent->{$CurName} = << "END";
$DocStruct->{'DECLARATION'}
$DocStruct->{'DESCRIPTION'}
$DocStruct->{'EXAMPLE'}
$DocStruct->{'ARGUMENT'}
$DocStruct->{'EXCEPTION'}
$DocStruct->{'RETURN'}
$DocStruct->{'SEE'}
END
		$DocStruct = {};
		$Headings->{$CurName} = $HeadingCounter++;
	}

	# インデックス
	if ($DocStruct->{'TITLE'}) {
		$FileContent->{'_'} = << "END";
TITLE 吉里吉里プラグイン $PlugInName ： $TitleName
$DocStruct->{'ABSTRACT'}
END
	}
}


sub asTDS {
	my ($text, $t) = @_;
	my $te = '';
	$te = "/$t" if ($t);
	$text =~ s|\n<<TDS|\n$te|g;
	$text =~ s|^<<TDS|$te|g;
	$text =~ s|\n>>|\n$t|g;
	return $text;
}


sub asTDSIL {
	my ($text) = @_;

# @{link method arg disp}
	while ($text =~ /(.)\@\{link\s+(.*?)\}/sg) {
		my ($pcr, $a) = ($1, $2);
		my $link = createLink($a);
		if ($link) {
			$pcr = ($pcr eq "\n") ? "\n" : "$pcr\n";
			$text =~ s/.\@\{link\s+(.*?)\}/${pcr}LINK ${link}\n/s;
		}
	}

# @{INLINE params}
# ->
# INLINE params
	while ($text =~ /(.)\@\{(.*?)\}/sg) {
		my ($pcr, $a) = ($1, $2);
		$pcr = ($pcr eq "\n") ? "\n" : "$pcr\n";
		$text =~ s/.\@\{(.*?)\}/${pcr}$a\n/s;
	}

	return $text;
}


# リンク作成
sub createLink {
	my $rv = '';
	my ($a) = @_;
	my @as = split(/\s+/, $a, 3);
	if (scalar(@as) >= 2) {
		my $ac;
		my $exfile = '';
		my $methodOmit = 0;
		my $argOmit = 0;

		# オンラインマニュアル
		my $online = 0;

		# 外部ファイル参照
		if ($as[0] =~ /^.*::/) {
			$as[0] =~ s/^(.*):://;
			$exfile = $1;
			$exfile =~ s/^\s*//;
			if (exists($OnlineRoot{$exfile})) {
				$online = $exfile;
				$exfile = $OnlineRoot{$exfile};
			}
		}

		# クラス＋メソッド省略時
		if ($as[0] eq '-') {
			$as[0] = $CurName;
			$methodOmit = 1;
		}
		# 引数名省略時
		if ($as[1] eq '-') {
			$ac = '';
			$argOmit = 1;
		# 引数名あり
		} else {
			$ac = getArgID($as[0], $as[1]);
		}
		my $ds = $as[2];
		# 表示文字列省略時
		if ($ds eq '') {
			$ds = $argOmit ? $as[0] : $as[1];
		}
		unless ($exfile) {
			$exfile = getFileNameBase($as[0]) . '.html';
		}
		my $url = ${exfile};
		$url .= "#${ac}" if ($ac);

		# 命名規則: URL = $TVPDocRoot/f_CLASS_METHOD.html
		if ($online eq '!TVP') {
			my $a = $as[0];
			$a =~ s/\./_/;
			$url = "${exfile}/f_${a}.html";
		} elsif ($online eq '!TJS') {
		# 命名規則: URL = $TJSDocRoot/dictionary.html
			my $a = lc($as[0]);
			$a =~ s/\..*//;
			$url = "${exfile}/${a}.html";
		}

		$rv = "${url} ${ds}";
	}

	return $rv;
}



sub removeEmptyPair {
	my ($text, $t) = @_;
	$text =~ s|\n$t\n+\/$t\n|\n|sg;
	return $text;
}

sub replaceCR {
	my ($text, $t) = @_;
	my $te = '';
	$te = "/$t" if ($t);
	$text =~ s/^\s*//;
	$text =~ s|\n<<CR>>|\n$te\n$t|g;
	$text =~ s|^<<CR>>|$te\n$t|g;
	return $text;
}


sub getArgID {
	my ($name, $arg) = @_;
	return "$name.$arg";
}

sub getMethodID {
	my ($name) = @_;
	return "$name";
}



sub getArgList {
	my ($texts) = @_;
	my $rv = '';
	my $dd = 0;
	my $hd = 0; # here doc

	foreach my $e (@$texts) {
		if ($e =~ /^\t*<<TDS$/) {
			$hd = 1;
			next;
		}

		if ($e =~ /^\t*>>$/) {
			$hd = 0;
			next;
		}

		if ($hd) {
			$e =~ s/^\t*//;
			$rv .= "$e\n";

		} elsif ($e =~ /^\t/) {
			unless ($dd) {
				$rv .= "DD+\n";
				$dd = 1;
			}
			$e =~ s/^\s*//;
			$e = replaceCR($e, 'DD');
			$rv .= "$e\n";

		} else {
			if ($dd) {
				$rv .= "/DD\n";
				$dd = 0;
			}
			my @es = split(/\s+/, $e, 3);
			$es[1] = "($es[1])";
			if ($es[2]) {
				if ($es[2] =~ /^\[(.*)\]$/) {
					$es[2] = "[default: $1]";
				}
			}
			$e = join(' ', @es);
			my $id = getArgID($CurName, $es[0]);
			$rv .= "DT+ id='$id'\n$e\n/DT\n";
		}
	}

	if ($dd) {
		$rv .= "/DD\n";
		$dd = 0;
	}

	return $rv;
}

