
#
# DLLプラグイン用
#

use strict;

use utf8;
my $ENCODE_NAME = ':encoding(shift_jis)';
binmode(STDIN,  $ENCODE_NAME);
binmode(STDOUT, $ENCODE_NAME);

my $CLASS_DEF = '';
my $EXTERN_VARS = '';
my %buf = ();
my $USE_DH = $ARGV[0] ? 1 : 0;	# データホルダ使用？

my $DefineID = genDefineID();

shift @ARGV;	# 残りはinclude
my $EXTINC = join('', map { qq|#include "$_"\n| } @ARGV);


my $temp_classdef = << 'END';
// Function : ###FUNC_NAME###
class ###CLASS_NAME### : public tTJSDispatch {
public:
	TVP_PLUGIN_FUNCCALL_DECL_ICD;
###CLASS_IMPL###
};

END

# read-onlyにならないなあ…むぅ。
my $temp_classdef_r = << 'END';
// Property (Read-only) : ###FUNC_NAME###
class ###CLASS_NAME### : public tTJSDispatch {
protected:
	tTJSVariant prop_value;

public:
	###CLASS_NAME###() : tTJSDispatch() {
		this->prop_value = ###PROP_VALUE###;
	}

	TVP_PLUGIN_PROPGET_DECL_ICD
	{
		if (membername) return TJS_E_MEMBERNOTFOUND;
		if (result) *result = this->prop_value;
		return TJS_S_OK;
	}

	TVP_PLUGIN_PROPSET_DECL_ICD
	{
		return TJS_E_NOTIMPL;
	}

###CLASS_IMPL###
};

END

my $temp_classdef_rw = << 'END';
// Property (Read-Write) : ###FUNC_NAME###
class ###CLASS_NAME### : public tTJSDispatch {
protected:
	tTJSVariant prop_value;

public:
	###CLASS_NAME###() : tTJSDispatch() {
		this->prop_value = ###PROP_VALUE###;
	}

	TVP_PLUGIN_PROPGET_DECL_ICD
	{
		if (membername) return TJS_E_MEMBERNOTFOUND;
		if (result) *result = this->prop_value;
		return TJS_S_OK;
	}

	TVP_PLUGIN_PROPSET_DECL_ICD
	{
		if (membername) return TJS_E_MEMBERNOTFOUND;
		if (param) this->prop_value = *param;
		return TJS_S_OK;
	}
###CLASS_IMPL###
};

END


my $TypeFunc = 1;
my $TypePropR  = 2;
my $TypePropW  = 4;
my $TypePropRW = $TypePropR | $TypePropW;


my ($a, $f, $impl);

my @Data = ();


while (<STDIN>) {
	chomp;
	push @Data, $_;
}

foreach (@Data) {
	# 続き
	if ($_ =~ /^::(\S+)/s) {
		$a = $1;
		$impl = '';
		while (<STDIN>) {
			last if ($_ =~ /^;;/s);
			$impl .= $_;
		}
		$buf{$a} =~ s/###CLASS_IMPL###/${impl}/gs;

	} else {
		chomp;
		$_  =~ s/^\s+//;
		next if (!$_ || $_ =~ /^#/);
		($a, $f) = split(/\s+/, $_);
		my $ptype = $TypeFunc;
		my $pvalue = '(tTVInteger)0';

		# クラス
		if ($f =~ s/\@class://) {
			# コールバック関数自動登録
			push @Data, "${a}WhenCreateNewFunction ${f}._WhenCreateNew";
			push @Data, "${a}WhenInvalidateFunction ${f}._WhenInvalidate";
			next;
		}

		# スタティック
		if ($f =~ s/\@static//) {
			# no action
		}

		# プロパティ： 
		if ($f =~ s/:([rw]+)//) {
			$ptype = $TypePropR  if ($1 eq 'r');
			$ptype = $TypePropRW if ($1 eq 'rw');
		}

		# プロパティ： 値
		if ($f =~ s/:(.+)//) {
			$pvalue = $1;
		}

		if ($a && $f) {
			my $t = 
				($ptype == $TypeFunc) ? $temp_classdef : (
				($ptype == $TypePropR) ? $temp_classdef_r : $temp_classdef_rw);
			$t =~ s/###CLASS_NAME###/$a/gs;
			$t =~ s/###FUNC_NAME###/$f/gs;
			$t =~ s/###PROP_VALUE###/$pvalue/gs;
			$buf{$a} = $t;

			my $v = $a;
			$v =~ s/^t//;
			$EXTERN_VARS .= "extern $a * $v;\n";
		}
	}
}


while (my ($k, $v) = (each %buf)) {
	$v =~ s/###CLASS_IMPL###//gs;

	$CLASS_DEF .= $v;
}


my $temp_base = << "END";
//---------------------------------------------------------------------------
// Generated by $0
//---------------------------------------------------------------------------
#ifndef ${DefineID}
#define ${DefineID}

END

# use holder ?
if ($USE_DH) {
	$temp_base .= << 'END';
#include "TableHolder.h"

extern CTableHolder * TableHolder;

END
}


$temp_base .= << "END";

$EXTINC
#include "TVPPlugIn.h"

${CLASS_DEF}
${EXTERN_VARS}
#endif ${DefineID}
//---------------------------------------------------------------------------
END

print $temp_base;

sub genDefineID {
	my $a = $0;
	$a =~ s|\\|/|g;
	$a =~ s|.*/||;
	$a =~ tr/./_/;
	$a = uc($a);
	my $rv .= "_TVP_PLUGIN_AUTOGEN_${$}_" . time() . "_${a}_H_";
	return $rv;
}


