#
# ネイティブクラス変換ツール
#

use strict;

use utf8;
my $ENCODE_NAME = ':encoding(shift_jis)';
binmode(STDIN,  $ENCODE_NAME);
binmode(STDOUT, $ENCODE_NAME);

# 1: plugin mode; 0: native mode
my $PluginMode = $ARGV[1];
# Is current static ? 
my $StaticDecl = 0;
# Is current no instance class?
my $NoInstance = 0;

my $FileName = '';
my $CurrentPP = '';

my $TJSNC_PREFIX = $PluginMode ? 'TJSNC_' : 'tjse_NativeClassClass_';
my $TJSNI_PREFIX = $PluginMode ? 'TJSNI_' : 'tjse_NativeInstanceClass_';

my $TS_CONSTRUCTOR     = 'constructor';
my $TS_EMPTY_FINALIZE  = 'emptyfinalize';
my $TS_EMPTY_CONSTRUCTOR = 'emptyconstructor';
my $TS_PROPERTY = 'property';
my $TS_METHOD   = 'method';
my $TS_GETTER   = 'getter';
my $TS_SETTER   = 'setter';
my $TS_BEGIN    = 'begin';
my $TS_END      = 'end';
my $TS_FILE     = 'file';
my $TS_COMMENT  = 'comment';

my $TS_STATIC_DECL = 'static';
my $TS_NO_INSTANCE = 'noinstance';
my $TS_READONLY = 'readonly';

my $TS_SEP = '//' . ('-' x 64);

main();

sub main
{

	# 出力バッファ
	my @OUT = ();

	my @RMode = ();
	my $Class = '';
	my $Member = '';
	my $Line = 1;

#	$CurrentPP = qq(#line 1);
#	push @OUT, "${CurrentPP}\n";

	while (<STDIN>) {
		chomp;
		my $a = $_;
		$Line++;

		$CurrentPP = qq(#line $Line "$FileName");

		if ($a =~ /^\@${TS_COMMENT}/) {
			next;
		}

		if ($a =~ /^\@${TS_FILE}\s+(\S+)/) {
			$FileName = $1;
			$CurrentPP = qq(#line $Line "$FileName");
			push @OUT, "${CurrentPP}";
			next;
		}


		# end of block
		if (@RMode && ($a eq '@;')) {
			my $mode = pop @RMode;
			my $o = '';

			if ($mode eq $TS_CONSTRUCTOR) {
				$o = getConstructorDeclEnd($Class);
				$StaticDecl = 0;

			} elsif ($mode eq $TS_METHOD) {
				$o = getMethodDeclEnd($Class, $Member);
				$StaticDecl = 0;

			} elsif ($mode eq $TS_PROPERTY) {
				$o = getPropDeclEnd($Class, $Member);
				$StaticDecl = 0;

			} elsif ($mode eq $TS_GETTER) {
				$o = getGetterDeclEnd($Class, $Member);

			} elsif ($mode eq $TS_SETTER) {
				$o = getSetterDeclEnd($Class, $Member);
			}
			push @OUT, $o;

			next;
		}

		if ($a =~ /^\@(\S+)(.*)/) {
			my ($mode, $fullparam) = ($1, $2);
			$fullparam =~ s/^\s*//;
			my @params = split(/\s+/, $fullparam);
			my $param = $params[0];
			my $parent = @RMode ? $RMode[$#RMode] : '';

#print STDERR "$mode, $parent, $param\n";

			checkNest($mode, $parent, $Line, $param);

			if ($mode eq $TS_END) {
				@RMode = ();
				push @OUT, getClassCreatorEnd($Class);
				$NoInstance = 0;
				next;

			} elsif ($mode eq $TS_CONSTRUCTOR) {
				push @RMode, $mode;
				$StaticDecl = ($param eq $TS_STATIC_DECL);
				push @OUT, getConstructorDeclBegin($Class);
				next;

			} elsif ($mode eq "${TS_EMPTY_FINALIZE};") {
				# 一行で完結するのでスタックには積まない
				push @OUT, getEmptyFinalizeDecl($Class);
				next;

			} elsif ($mode eq "${TS_EMPTY_CONSTRUCTOR};") {
				# 一行で完結するのでスタックには積まない
				push @OUT, getEmptyConstructorDecl($Class);
				next;

			} elsif ($mode eq $TS_METHOD) {
				push @RMode, $mode;
				$Member = $param;
				$StaticDecl = ($params[1] eq $TS_STATIC_DECL);
				push @OUT, getMethodDeclBegin($Class, $Member);
				next;

			} elsif ($mode eq $TS_PROPERTY) {
				# property NAME [static] [readonly TYPE VALUE]
				# TYPE = INT|REAL|OBJ|VAR|STR|PTR
				$Member = $param;
				$StaticDecl = ($params[1] eq $TS_STATIC_DECL);
				my $ro = ($params[$StaticDecl ? 2 : 1] eq $TS_READONLY);
				if ($ro) {
					my @tp = @params;
					shift @tp;
					shift @tp;
					shift @tp if ($StaticDecl);
					# 一行で完結するのでスタックには積まない
					my $type  = shift @tp;
					my $value = join(' ', @tp);
					push @OUT, getPropDeclReadOnly($Class, $Member, $type, $value);
				} else {
					push @RMode, $mode;
					push @OUT, getPropDeclBegin($Class, $Member);
				}
				next;

			} elsif ($mode eq $TS_GETTER) {
				if ($param eq 'deny;') {
					# 一行で完結するのでスタックには積まない
					push @OUT, getDenyGetterDecl($Class, $Member);
				} else {
					push @RMode, $mode;
					push @OUT, getGetterDeclBegin($Class, $Member);
				}
				next;

			} elsif ($mode eq $TS_SETTER) {
				if ($param eq 'deny;') {
					# 一行で完結するのでスタックには積まない
					push @OUT, getDenySetterDecl($Class, $Member);
				} else {
					push @RMode, $mode;
					push @OUT, getSetterDeclBegin($Class, $Member);
				}
				next;

			} elsif ($mode eq $TS_BEGIN) {
				$Class = $param;
				$NoInstance = ($params[1] eq $TS_NO_INSTANCE);
				push @RMode, $mode;
				push @OUT, getDefClassID($Class);
				push @OUT, getInstanceCreator($Class);
				push @OUT, getClassCreatorBegin($Class);
				next;

			} else {

				abort($Line, "Syntax error: '$a'?");
			}
		}

		push @OUT, $a;

	}

	if (@RMode) {
		abort($Line, "Nest error: unexpected EOF.");
	}

	print join("\n", @OUT, '');

}

sub abort
{
	my ($line, $errmes) = @_;
	die ("line ${line}: ${errmes}\n");
}

sub checkNest
{
	my ($mode, $parent, $line, $param) = @_;
	my $errmes = '';

	if (
		(	$mode eq $TS_END ||
			$mode eq $TS_CONSTRUCTOR ||
			$mode eq $TS_METHOD ||
			$mode eq $TS_PROPERTY
		) && $parent ne $TS_BEGIN )
	{
		$errmes = "'\@$mode' must be child of '$TS_BEGIN'. Current mode is '$parent'.";
	}

	if ($mode eq $TS_END && $parent ne $TS_BEGIN)
	{
		$errmes = "'\@$mode' must be pair with '$TS_BEGIN'. Current mode is '$parent'.";
	}

	if (
		(	$mode eq $TS_GETTER ||
			$mode eq $TS_SETTER
		) && ($parent ne $TS_PROPERTY) )
	{
		$errmes = "'\@$mode' must be child of '$TS_PROPERTY'. Current mode is '$parent'.";
	}

	if ($mode eq $TS_BEGIN && $parent) {
		$errmes = "'\@$TS_BEGIN' must be root.";
	}

	if ($errmes) {
		abort($line, $errmes);
	}
}



# 
sub getInstanceCreator
{
	my $c = shift;
	my $rv;

	# プラグイン
	if ($PluginMode) {
		$rv = << "END";
// インスタンスクリエイター
static iTJSNativeInstance * Create_${c}_NI()
{
	return new ${TJSNI_PREFIX}${c}();
}
END
	}

	return $rv;
}


# クラスIDの定義。ネイティブでは何もしない。
sub getDefClassID
{
	my $c = shift;
	my $rv;

	# プラグイン
	if ($PluginMode) {
		$rv = << "END";
$TS_SEP
// TJS Native Class '${c}'
$TS_SEP
/*
  Native Instance Class   : ${TJSNI_PREFIX}${c}
  Native Instance Creator : Create_${c}_NI()
  Native Class Creator    : Create_${c}_NC()
  Generated by $0
*/
#define TJS_NATIVE_CLASSID_NAME ClassID_${TJSNI_PREFIX}${c}
static tjs_int32 TJS_NATIVE_CLASSID_NAME = -1;
END

	}
	return $rv;
}


# Class creator
sub getClassCreatorBegin
{
	my $c = shift;
	my $rv;

	# プラグイン
	if ($PluginMode) {
		$rv = << "END";
// クラスクリエイター
iTJSDispatch2 * Create_${c}_NC()
{
tTJSNativeClassForPlugin * classobj =
	TJSCreateNativeClassForPlugin(TJS_W("${c}"), Create_${c}_NI);

/*
このマクロと TJS_END_NATIVE_MEMBERS マクロで挟まれた場所に、クラスの
メンバとなるべきメソッドやプロパティの記述をします。
*/
// ネイティブメンバ定義部ここから
TJS_BEGIN_NATIVE_MEMBERS(/*TJS class name*/${c})
${CurrentPP}
END

	# ネイティブ
	} else {
		my $ci = ($NoInstance) ? "return NULL;" : "return new ${TJSNI_PREFIX}${c}();";
		$rv = << "END";
$TS_SEP
// begin TJS Native Class '${c}'
//   Native Instance Class: ${TJSNI_PREFIX}${c}
// 
$TS_SEP
// ネイティブクラス ${c} コンストラクタ
tTJSNativeInstance * ${TJSNC_PREFIX}${c}::CreateNativeInstance()
{
	${ci}
}

// クラスID
${CurrentPP}
tjs_uint32 ${TJSNC_PREFIX}${c}::ClassID = (tjs_uint32)-1;

// ネイティブクラス ${c} のコンストラクタの実装
${TJSNC_PREFIX}${c}::${TJSNC_PREFIX}${c}() : tTJSNativeClass(TJS_W("${c}")) \
{ \
TJS_BEGIN_NATIVE_MEMBERS(${c}) \
${CurrentPP}
END
	}

	return $rv;
}


sub getClassCreatorEnd
{
	my $c = shift;
	my $rv;

	# プラグイン
	if ($PluginMode) {
		$rv = << "END";
// ネイティブメンバ定義部ここまで
TJS_END_NATIVE_MEMBERS

	/*
		この関数は classobj を返します。
	*/
	return classobj;
}

// 念のため
#undef TJS_NATIVE_CLASSID_NAME

$TS_SEP
// end of TJS Native Class '${c}'
$TS_SEP

END

	# ネイティブ
	} else {
		$rv = << "END";
	TJS_END_NATIVE_MEMBERS \
}
$TS_SEP
// end of TJS Native Class '${c}'
$TS_SEP
END
	}

	return $rv;
}

# Empty Finalize
sub getEmptyFinalizeDecl
{
	my $c = shift;
	my $rv = << "END";
$TS_SEP
// empty finalize method
$TS_SEP
TJS_DECL_EMPTY_FINALIZE_METHOD
END
	return $rv;
}

# Empty Constructor
sub getEmptyConstructorDecl
{
	my $c = shift;
	my $rv = getConstructorDeclBegin($c) . "\treturn TJS_S_OK;\n" . getConstructorDeclEnd($c);
	return $rv;
}


# Constructor
sub getConstructorDeclBegin
{
	my $c = shift;
	my $niflag = $NoInstance ? '_NO_INSTANCE' : '';
	my $macroarg;
	my $rv;

	if ($niflag) {
		$macroarg = "/*TJS class name*/${c}";
	} else {
		$macroarg = << "END";
	/*var.name*/_this,
	/*var.type*/${TJSNI_PREFIX}${c},
	/*TJS class name*/${c}
END
	}

	$rv = << "END";
$TS_SEP
// constructor ${c}()
$TS_SEP
TJS_BEGIN_NATIVE_CONSTRUCTOR_DECL${niflag}(${macroarg})
{
${CurrentPP}
END

	return $rv;
}

sub getConstructorDeclEnd
{
	my $c = shift;
	my $stflag = $StaticDecl ? 'STATIC_' : '';

	my $rv = << "END";
}
TJS_END_NATIVE_${stflag}CONSTRUCTOR_DECL(/*TJS class name*/${c})
$TS_SEP
// end of constructor ${c}()
$TS_SEP
END
	return $rv;
}


# Method
sub getMethodDeclBegin
{
	my $c = shift;
	my $m = shift;
	my $getni = $NoInstance ? '// no instance class' : "TJS_GET_NATIVE_INSTANCE(/*var. name*/_this, /*var. type*/${TJSNI_PREFIX}${c});";

	my $rv = << "END";
$TS_SEP
// method ${c}.${m}()
$TS_SEP
TJS_BEGIN_NATIVE_METHOD_DECL(/*func. name*/${m})
{
	void * membername = NULL;	// DUMMY
	${getni}
${CurrentPP}
END
	return $rv;
}

sub getMethodDeclEnd
{
	my $c = shift;
	my $m = shift;
	my $stflag = $StaticDecl ? 'STATIC_' : '';

	my $rv = << "END";
}
TJS_END_NATIVE_${stflag}METHOD_DECL(/*func. name*/${m})
$TS_SEP
// end of method ${c}.${m}()
$TS_SEP
END
	return $rv;
}


# Property
sub getPropDeclBegin
{
	my $c = shift;
	my $m = shift;
	my $stflag = $StaticDecl ? 'static ' : '';

	my $rv = << "END";
$TS_SEP
// ${stflag}property ${c}.${m}
$TS_SEP
TJS_BEGIN_NATIVE_PROP_DECL(/*prop. name*/${m})
{
END
	return $rv;
}

sub getPropDeclEnd
{
	my $c = shift;
	my $m = shift;
	my $stflag = $StaticDecl ? 'STATIC_' : '';

	my $rv = << "END";
}
TJS_END_NATIVE_${stflag}PROP_DECL(/*prop. name*/${m})
$TS_SEP
// end of property ${c}.${m}
$TS_SEP
END
	return $rv;
}




# Property getter
sub getGetterDeclBegin
{
	my $c = shift;
	my $m = shift;
	my $getni = $NoInstance ? '// no instance class' : "TJS_GET_NATIVE_INSTANCE(/*var. name*/_this, /*var. type*/${TJSNI_PREFIX}${c});";

	my $rv = << "END";
$TS_SEP
// getter ${c}.${m}
$TS_SEP
TJS_BEGIN_NATIVE_PROP_GETTER
{
	${getni}
${CurrentPP}
END
	return $rv;
}

sub getGetterDeclEnd
{
	my $c = shift;
	my $m = shift;
	my $rv = << "END";
}
TJS_END_NATIVE_PROP_GETTER
$TS_SEP
// end of getter ${c}.${m}
$TS_SEP
END
	return $rv;
}


# Setter
sub getSetterDeclBegin
{
	my $c = shift;
	my $m = shift;
	my $getni = $NoInstance ? '// no instance class' : "TJS_GET_NATIVE_INSTANCE(/*var. name*/_this, /*var. type*/${TJSNI_PREFIX}${c});";

	my $rv = << "END";
$TS_SEP
// setter ${c}.${m}
$TS_SEP
TJS_BEGIN_NATIVE_PROP_SETTER
{
	${getni}
${CurrentPP}
END
	return $rv;
}

sub getSetterDeclEnd
{
	my $c = shift;
	my $m = shift;
	my $rv = << "END";
}
TJS_END_NATIVE_PROP_SETTER
$TS_SEP
// end of setter property ${c}.${m}
$TS_SEP
END
	return $rv;
}


# Getter (deny)
sub getDenyGetterDecl
{
	my $c = shift;
	my $m = shift;
	my $rv = << "END";
$TS_SEP
// getter ${c}.${m} is denied
$TS_SEP
TJS_DENY_NATIVE_PROP_GETTER
END
	return $rv;
}


# Setter (deny)
sub getDenySetterDecl
{
	my $c = shift;
	my $m = shift;
	my $rv = << "END";
$TS_SEP
// setter ${c}.${m} is denied
$TS_SEP
TJS_DENY_NATIVE_PROP_SETTER
END
	return $rv;
}


# Property (readonly)
sub getPropDeclReadOnly
{
	my $c = shift;
	my $m = shift;
	my $type  = shift;
	my $value = shift;

	my $stflag = $StaticDecl ? 'static ' : '';

	my $rv = getPropDeclBegin($c, $m);
	$rv .= getGetterDeclBegin($c, $m);
	$rv .= << "END";
	TJSE_SET_RETURN_VALUE_${type}(${value})
	return TJS_S_OK;
END
	$rv .= getGetterDeclEnd($c, $m);
	$rv .= getDenySetterDecl($c, $m);
	$rv .= getPropDeclEnd($c, $m);

	return $rv;
}


