
#
# DLLプラグイン用
#

use utf8;
my $ENCODE_NAME = ':encoding(shift_jis)';
binmode(STDIN,  $ENCODE_NAME);
binmode(STDOUT, $ENCODE_NAME);

my $REGISTER = '';
my $RELEASE = '';
my $VARS = '';
my $HNAME = "$ARGV[0].h" || 'util_geom.h';
my $USE_DH = $ARGV[1] ? 1 : 0;	# データホルダ使用？
my %ICO_CLASS = ();

my @RegisterList = ();
my @ReleaseList = ();


# グローバル関数
my $temp_register = << 'END';
	//-----------------------------------------------------------------------
	###VAR_NAME### = new ###CLASS_NAME###();
	val = tTJSVariant(###VAR_NAME###);
	###VAR_NAME###->Release();
	global->PropSet(TJS_MEMBERENSURE ###CALL_FLAG###, TJS_W("###FUNC_NAME###"), NULL, &val, global);
	//-----------------------------------------------------------------------
END

# メソッド
my $temp_register_wc = << 'END';
	//-----------------------------------------------------------------------
	###VAR_NAME### = new ###CLASS_NAME###();
	val = tTJSVariant(###VAR_NAME###);
	###VAR_NAME###->Release();
	classdef_###ICO_CLASS###->PropSet(TJS_MEMBERENSURE ###CALL_FLAG###, TJS_W("###FUNC_NAME###"), NULL, &val, classdef_###ICO_CLASS###);
	//-----------------------------------------------------------------------
END

# クラス定義取得
my $temp_register_ico = << 'END';
	// ###ICO_CLASS### オブジェクト（クラス定義）を取得
	global->PropGet(TJS_MEMBERMUSTEXIST, TJS_W("###ICO_CLASS###"), NULL, &ico_result, global);
	iTJSDispatch2 * classdef_###ICO_CLASS### = ico_result.AsObjectNoAddRef();
END

my $temp_release = << 'END';
		global->DeleteMember(0, TJS_W("###FUNC_NAME###"), NULL, global);
END

my $temp_release_wc = << 'END';
		if (classdef_###ICO_CLASS###) {
			classdef_###ICO_CLASS###->DeleteMember(0, TJS_W("###FUNC_NAME###"), NULL, classdef_###ICO_CLASS###);
		}
END

my @Data = ();


while (<STDIN>) {
	chomp;
	push @Data, $_;
}

foreach (@Data) {
	$_  =~ s/^\s+//;
	next if (!$_ || $_ =~ /^#/);
	last if ($_ =~ /^::/);
	my ($a, $f) = split(/\s+/, $_);
	my $cf = '';

	if ($a && $f) {
		my $cppvar = "v${a}Instance";
		# スタティック
		if ($f =~ s/\@static//) {
			$cf = '| TJS_STATICMEMBER';
		}

		if ($f =~ s/:([rw]+).*//) {
			# プロパティ；処理不要
		}

		my $t;
		my $ico = '';

		# class def.
		my $classfile = '';
		if ($f =~ s/\@class:(.+)//) {
			$classfile = $1;
			my $ud;
			$t = genClassdef($classfile, $f);
			push @RegisterList, $t; # 登録はリストに
			# コールバック関数自動登録
			push @Data, "${a}WhenCreateNewFunction ${f}._WhenCreateNew";
			push @Data, "${a}WhenInvalidateFunction ${f}._WhenInvalidate";

		# function | property
		} else {
			{
				# コンテキスト指定？
				my ($a, $b) = split(/\./, $f, 2);
				if ($b) {
					($ico, $f) = ($a, $b);
				}
			}

			if ($ico && !exists($ICO_CLASS{$ico})) {
				$t = $temp_register_ico;
				$t =~ s/###ICO_CLASS###/$ico/gs;
				push @RegisterList, $t; # 登録はリストに
				$RELEASE .= $t; # 解放は直に

				$ICO_CLASS{$ico} = 1;
			}

			$VARS .= "$a * $cppvar;\n";

			$t = $ico ? $temp_register_wc : $temp_register;
			$t =~ s/###ICO_CLASS###/$ico/gs;
			$t =~ s/###CLASS_NAME###/$a/gs;
			$t =~ s/###FUNC_NAME###/$f/gs;
			$t =~ s/###VAR_NAME###/$cppvar/gs;
			$t =~ s/###CALL_FLAG###/$cf/gs;
			push @RegisterList, $t; # 登録はリストに

		}
		$t = $ico ? $temp_release_wc : $temp_release;
		$t =~ s/###ICO_CLASS###/$ico/gs;
		$t =~ s/###FUNC_NAME###/$f/gs;
		push @ReleaseList, $t; # 解放もリストに

	}
}

$REGISTER .= join('', @RegisterList);
$RELEASE  .= join('', reverse @ReleaseList); # 解放は逆順



my $temp_base = << "END";
//---------------------------------------------------------------------------
// Generated by $0
//---------------------------------------------------------------------------
#include "${HNAME}"
END

# use holder ?
if ($USE_DH) {
	$temp_base .= << 'END';
#include "TableHolder.h"
#include <lmcons.h>

//---------------------------------------------------------------------------

CTableHolder * TableHolder;

END
} else {
	$temp_base .= << 'END';
#include <lmcons.h>

//---------------------------------------------------------------------------

END
}

$temp_base .= << "END";

void TVPCallBeforeV2LinkExit(void);

${VARS}

//---------------------------------------------------------------------------
#pragma argsused
int WINAPI DllEntryPoint(HINSTANCE hinst, unsigned long reason, void* lpReserved)
{
	return 1;
}
//---------------------------------------------------------------------------
static tjs_int GlobalRefCountAtInit = 0;
extern "C" HRESULT _stdcall _export V2Link(iTVPFunctionExporter *exporter)
{
	TVPInitImportStub(exporter);

	tTJSVariant val;
	tTJSVariant ico_result;
	iTJSDispatch2 * global = TVPGetScriptDispatch();

END

# use holder ?
if ($USE_DH) {
	$temp_base .= << "END";
	//-----------------------------------------------------------------------
	// データホルダー
	TableHolder = new CTableHolder();
	TableHolder->createTable(DEFAULT_ANGLE_ACC, DEFAULT_SIN_ACC);
	//-----------------------------------------------------------------------
END
}

$temp_base .= << "END";
${REGISTER}

	global->Release();
	val.Clear();
	GlobalRefCountAtInit = TVPPluginGlobalRefCount;

	TVPCallBeforeV2LinkExit();

	return S_OK;
}
//---------------------------------------------------------------------------
extern "C" HRESULT _stdcall _export V2Unlink()
{
	if(TVPPluginGlobalRefCount > GlobalRefCountAtInit) return E_FAIL;
	iTJSDispatch2 * global = TVPGetScriptDispatch();
	tTJSVariant ico_result;

	if (global) {
${RELEASE}

END

# use holder ?
if ($USE_DH) {
	$temp_base .= << "END";
		// データホルダー
		delete TableHolder;
END
}

$temp_base .= << "END";
	}

	if (global) global->Release();
	TVPUninitImportStub();

	return S_OK;
}
//---------------------------------------------------------------------------
static tjs_char option_message[] = L"";
//---------------------------------------------------------------------------
extern "C" const wchar_t* _stdcall _export GetOptionDesc()
{
	return option_message;
}
//---------------------------------------------------------------------------
END

print $temp_base;


sub genClassdef {
	my ($file, $name) = @_;
	my $tjs = '';
	my $crnew = '_WhenCreateNew';
	my $invld = '_WhenInvalidate';
	my $ud = 0;

	open (FT, $file) or die ("$file: $!");
	binmode(FT,  $ENCODE_NAME);
	while (<FT>) {
		chomp;
		my $x = $_;
		$x =~ s|/\*WHEN_NEW\*/|$crnew\(\.\.\.\);|;
		$x =~ s|/\*WHEN_INVALIDATE\*/|$invld\(\);|;
		$tjs .= qq|\t\tL"$x\\n"\n|;
	}
	close FT;
	my $rv = << "END";
	//-----------------------------------------------------------------------
	// Define class $name
	TVPExecuteExpression(
$tjs
		, NULL);
	//-----------------------------------------------------------------------
END

	return $rv;
}

