
# Makefileを作るためのスクリプト

use utf8;
my $ENCODE_NAME = ':encoding(shift_jis)';
binmode(STDIN,  $ENCODE_NAME);
binmode(STDOUT, $ENCODE_NAME);

# 基本セット
my @Targets = qw(
generic
graph
sqlite
);
#socket


# ソース追加分
my @ExtTargets = qw(
sqliteplus
sqlite3
);


my $TargetList = join(' ', @Targets);
my @Out = ();


# --- MEMO
push @Out, << "END";
# -----------------------------
# Generated by $0
# -----------------------------

!if !$d(KRKR_PLUGIN_CONFIG_MAK)
!include Config.mak
!endif

END

# --- include
#push @Out, '!include Config.mak';
push @Out, '';

# --- all
push @Out, "all: common $TargetList";

# --- make
# Makefile自身がターゲット
push @Out, << 'END';
make:
	$(PERL) tool/makemake.pl > Makefile
END

# --- target
foreach my $e ('common', @Targets) {
push @Out, << "END";
$e: revision
	cd \$\@
	\$(MAKE)
	\@cd ..

END
}

# --- install
push @Out, makeRule('install', @Targets);
# --- clean
push @Out, makeRule('clean', 'common', @Targets) . "\n";
# --- src_clean
push @Out, makeRule('src_clean', 'common', @Targets);

# --- doc
{
my $docs;
my $docclean = "doc_clean:\n";
my $tgdoc = 'doc:';
foreach my $e (@Targets) {
	my $ef = "util_$e";
	my $html = "\$(HTMLMAN_DIR)\\$ef\\index.html";
	my $tdf  = "\$(TDFMAN_DIR)\\$ef\\index.tdf";
	my $cmd = q{-for /F %e in ('dir /B $(TDFMAN_DIR)\\} . $ef . q{\*.tdf') do  } ;
	$cmd .= q{$(TDF2HTML) -s $(SKELETON_FILE) $(TDFMAN_DIR)\\} . $ef . q{\%e > $(HTMLMAN_DIR)\\} . $ef . q{\%~ne.html};

	$tgdoc .= " $html";
$docs .= << "END";
$tdf: \$(MAN_DIR)\\$ef.txt
	-mkdir \$(TDFMAN_DIR)\\$ef
	\$(PERL) tool/makedoc.pl < \$(MAN_DIR)\\$ef.txt
$html: $tdf
	-mkdir \$(HTMLMAN_DIR)\\$ef
	$cmd

END

	$docclean .= << "END";
	-\$(RM) \$(TDFMAN_DIR)\\$ef\\*.tdf
	-\$(RM) \$(HTMLMAN_DIR)\\$ef\\*.html
END
}

push @Out, $docs, $docclean, $tgdoc;
}


# --- release
# リリース用：べた書き
{

# バイナリアーカイブ
my @RelFiles = qw(
CHANGES
*.txt
doc/html
doc/style
doc/images
lib/sqlite3.dll
);

# ソースアーカイブ
my @SrcFiles = qw(
src/common/Makefile
src/common/readme.txt
src/common/md5.*
src/common/NIHolder.*
src/common/OSVersionInfo.*
src/common/GflSHA.*
src/common/TableHolder.*
src/common/TPMT19937ar.*
src/common/TVPPlugIn.h
src/tool/*.pl
src/Makefile
src/Config.mak
src/Rules.mak
src/forms.txt
src/readme.txt
lib/dummy.txt
);

# デモアーカイブ
my @DemoFiles = qw(
demo/image
demo/scenario/*.ks
demo/scenario/*.tjs
demo/system/Override.tjs
demo/startup.tjs
demo/readme.txt
);


my @RelSrcSuffix = qw(
h
c
cpp
tjs
txt
def
pl
);

foreach my $e (@Targets) {
	my $dd = "src/${e}";
	push @RelFiles, "lib/util_${e}.dll";
	push @SrcFiles, "${dd}/Makefile";
	foreach my $e2 (@RelSrcSuffix) {
		push @SrcFiles, "${dd}/*.${e2}";
	}
}

foreach my $e (@ExtTargets) {
	my $dd = "src/${e}";
	push @SrcFiles, "${dd}/Makefile";
	foreach my $e2 (@RelSrcSuffix) {
		push @SrcFiles, "${dd}/*.${e2}";
	}
}




my $relfls  = join(' ', @RelFiles);
my $srcfls  = join(' ', @SrcFiles);
my $demofls = join(' ', @DemoFiles);
push @Out, << "END";
release: all src_clean doc_clean doc release_only

release_only:
	\@cd ..
	-\$(RM) \$(TP_REL_FILE)
	-\$(RM) \$(TP_REL_SRC_FILE)
	-mzp -r \$(TP_REL_FILE) $relfls
	-mzp -r \$(TP_REL_SRC_FILE) $srcfls
	\@cd src
END

#	-\$(RM) \$(TP_REL_DEMO_FILE)
#	-mzp -r \$(TP_REL_DEMO_FILE) $demofls

}

# --- revision
push @Out, << 'END';
revision:
	$(PERL) tool/update_revision_h.pl $(TP_VERSION) $(TVP_VERSION)

END


# 書き出し
print join("\n", @Out);

sub makeRule {
	my ($t, @target) = @_;
	my $rc = "$t:\n";
	foreach my $e (@target) {
		$rc .= << "END";
	cd $e
	\$(MAKE) $t
	\@cd ..
END
	}

	return $rc;
}

